/**
 * @file geode_accel.c
 * @brief Adding hardware based acceleration
 * @author Juergen Beisert <jbeisert@netscape.net>
 *
 * This driver bases on Xorg's fbdev and the original nsc driver
 */

#undef DEBUG

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

/* all driver need this */
#include <xf86.h>
#include <xf86_OSproc.h>

/* Xfree86 header files */
#include <xaa.h>
#include <miline.h>
#include <xaarop.h>

#include "geode-fb.h"

/* Diese Optionen knnen gezielt einzelne Beschleunigungen abschalten
 * diese gehren in die "Device"-Section fr den Grafikhardware-Treiber!
 * Damit kann man sie abschalten, ohne den Treiber jedesmal ndern zu mssen!
 */
#if 0

"XaaNoScreenToScreenCopy",
"XaaNoSolidFillRect",
"XaaNoSolidFillTrap",
"XaaNoSolidTwoPointLine",
"XaaNoSolidBresenhamLine",
"XaaNoSolidHorVertLine",
"XaaNoDashedTwoPointLine",
"XaaNoDashedBresenhamLine",
"XaaNoMono8x8PatternFillRect",
"XaaNoMono8x8PatternFillTrap",
"XaaNoColor8x8PatternFillRect",
"XaaNoColor8x8PatternFillTrap",
"XaaNoCPUToScreenColorExpandFill",
"XaaNoScanlineCPUToScreenColorExpandFill",
"XaaNoScreenToScreenColorExpandFill",
"XaaNoImageWriteRect",
"XaaNoScanlineImageWriteRect",
"XaaNoWriteBitmap",
"XaaNoWritePixmap",
"XaaNoPixmapCache",
"XaaNoOffscreenPixmaps",

#endif

/*
 * Some curious effects:
 * With a scratch pad size of 3072 and a line offset of 4096 (SXGA with 64k colours)
 * the BLT engine cannot blit more than 1024 pixel in a line. All pixel above the
 * 1024th pixel is copied to the left beginning of the line.
 * With 4096 bytes of scratch pad RAM it works again.
 * The BLT function checks the line length. With 3072 scratch pad RAM its seems
 * no problem to copy up to 1536 pixel per line in one step. So 1280 should also
 * work, but it does not! Don't know why.
 */

/* to be supported ROPs
R = Result to be stored at Destination
D = Destination data (previous data) before operation
P = Pattern data
S = Source data
*/

/**
 * @brief what components are required for the XAAPatternROP
 *
 * ROPs for planemask = 0xFFFFFFFF:
 **/
static const int pattern_components[16] =
{
	[0] = 0,			/* 0x00 -> R = 0 */
	[1] = ROP_PAT | ROP_DST,	/* 0x5F -> R = !(D & P) */
	[2] = ROP_PAT | ROP_DST,	/* 0x50 -> R= P & !D */
	[3] = ROP_PAT,			/* 0xF0 -> R = P */
	[4] = ROP_PAT | ROP_DST,	/* 0x0A -> R = D & !P */
	[5] = ROP_DST,			/* 0xAA -> R = D */
	[6] = ROP_PAT | ROP_DST,	/* 0x5A -> R = D ^ P */
	[7] = ROP_PAT | ROP_DST,	/* 0xFA -> R = D | P */
	[8] = ROP_PAT | ROP_DST,	/* 0x05 -> R = !(D | P) */
	[9] = ROP_PAT | ROP_DST,	/* 0xA5 -> R = !(P ^ D) */
	[10] = ROP_DST,			/* 0x55 -> R = !D */
	[11] = ROP_PAT | ROP_DST,	/* 0xF5 -> R = P | !D */
	[12] = ROP_PAT,			/* 0x0F -> R = !P */
	[13] = ROP_PAT | ROP_DST,	/* 0xAF -> R = D | !P */
	[14] = ROP_PAT | ROP_DST,	/* 0x5F -> R = !(D & P) */
	[15] = 0			/* 0xFF -> R = 1 */
};

/**
 * @brief what components are required for the XAAPatternROP_PM
 *
 * ROPs for planemask != 0xFFFFFFFF:
 **/
static const int pattern_components_pm[16] =
{
	[0] = ROP_PAT | ROP_DST,		/* 0x0A -> R = D & !P */
	[1] = ROP_PAT | ROP_DST | ROP_SRC,	/* 0xA2 -> R = D & (P | !S) */
	[2] = ROP_PAT | ROP_DST | ROP_SRC,	/* 0x51 -> R = !(D | ( S & !P)) */
	[3] = ROP_PAT | ROP_DST | ROP_SRC,	/* 0xE2 -> R = D ^ (S & (P ^ D)) */
	[4] = ROP_PAT | ROP_DST | ROP_SRC,	/* 0x2A -> R = D & !(P & S) */
	[5] = ROP_DST,				/* 0xAA -> R = D */
	[6] = ROP_PAT | ROP_DST | ROP_SRC,	/* 0x6A -> R = D ^ (P & S) */
	[7] = ROP_PAT | ROP_DST | ROP_SRC,	/* 0xEA -> R = D | (P & S) */
	[8] = ROP_PAT | ROP_DST | ROP_SRC,	/* 0x15 -> R = !(D | (P & S)) */
	[9] = ROP_PAT | ROP_DST | ROP_SRC,	/* 0x95 -> R = !(D ^ (P & S)) */
	[10] = ROP_PAT | ROP_DST,		/* 0x5A -> R = D ^ P */
	[11] = ROP_PAT | ROP_DST | ROP_SRC,	/* 0xD5 -> R = !(D & !(P & S)) */
	[12] = ROP_PAT | ROP_DST | ROP_SRC,	/* 0x3A -> R = S ^(P | (D ^S)) */
	[13] = ROP_PAT | ROP_DST | ROP_SRC,	/* 0xAE -> R = D | (S & !P) */
	[14] = ROP_PAT | ROP_DST | ROP_SRC,	/* 0x5D -> R = !(D & (P | !S) */
	[15] = ROP_PAT | ROP_DST		/* 0xFA -> R = D | P */
};

/**
 * @brief what components are required for the XAACopyROP
 *
 * ROPs for planemask = 0xFFFFFFFF:
 **/
static const int copy_components[16] =
{
	[0] = 0,			/* 0x00 -> R = 0 */
	[1] = ROP_DST | ROP_SRC,	/* 0x88 -> S = D & S */
	[2] = ROP_DST | ROP_SRC,	/* 0x44 -> R = S & !D */
	[3] = ROP_SRC,			/* 0xCC -> R = S */
	[4] = ROP_DST | ROP_SRC,	/* 0x22 -> R = D & !S */
	[5] = ROP_DST,			/* 0xAA R = D */
	[6] = ROP_DST | ROP_SRC,	/* 0x66 -> R = D ^ S */
	[7] = ROP_DST | ROP_SRC,	/* 0xEE -> R = D | S */
	[8] = ROP_DST | ROP_SRC,	/* 0x11 -> R = !(D | S) */
	[9] = ROP_DST | ROP_SRC,	/* 0x99 -> R = !(D ^ S) */
	[10] = ROP_DST,			/* 0x55 -> R = !D */
	[11] = ROP_DST | ROP_SRC,	/* 0xDD -> R = S | !D */
	[12] = ROP_SRC,			/* 0x33 -> R = !S */
	[13] = ROP_DST | ROP_SRC,	/* 0xBB -> R = D | !S */
	[14] = ROP_DST | ROP_SRC,	/* 0x77 -> R = !(D & S) */
	[15] = 0,			/* 0xFF -> R = 1 */
};

/**
 * @brief what components are required for the XAACopyROP_PM
 *
 * ROPs for planemask != 0xFFFFFFFF:
 **/
static const int copy_components_pm[16] =
{
	[0] = 0,				/* 0x00 -> R = 0 */
	[1] = ROP_PAT | ROP_DST | ROP_SRC,	/* 0x8A -> R = D & (S | !P) */
	[2] = ROP_PAT | ROP_DST | ROP_SRC,	/* 0x45 -> R = !(D | (P & !S)) */
	[3] = ROP_PAT | ROP_DST | ROP_SRC,	/* 0xCA -> R = D ^ (P & (S ^ D)) */
	[4] = ROP_PAT | ROP_DST | ROP_SRC,	/* 0x2A -> R = D & !(P & S) */
	[5] = ROP_DST,				/* 0xAA -> R = D */
	[6] = ROP_PAT | ROP_DST | ROP_SRC,	/* 0x6A -> R = D ^ (P & S) */
	[7] = ROP_PAT | ROP_DST | ROP_SRC,	/* 0xEA -> R = D | (P & S) */
	[8] = ROP_PAT | ROP_DST | ROP_SRC,	/* 0x15 -> R = !(D | (P & S)) */
	[9] = ROP_PAT | ROP_DST | ROP_SRC,	/* 0x95 -> R = !(D  ^(P & S)) */
	[10] = ROP_DST,				/* 0x55 -> R = !D */
	[11] = ROP_PAT | ROP_DST | ROP_SRC,	/* 0xD5 -> R = !(D & !(P & S)) */
	[12] = ROP_PAT | ROP_DST | ROP_SRC,	/* 0x2E -> R = P ^ (S | (D ^ P))*/
	[13] = ROP_PAT | ROP_DST | ROP_SRC,	/* 0xBA -> R = D | (P & !S) */
	[14] = ROP_PAT | ROP_DST | ROP_SRC,	/* 0x75 -> R = !(D & (S | !P)) */
	[15] = 0,				/* 0xFF -> R = 1 */
};

/**
 * @brief write a ne gamma table into the gamma correction LUT
 * @param[in] fPtr FIXME
 * @param[in] colour Pointer to 256 RGB entries
 *
 * Note: This function will only be used from the xv extension.
 *
 * @return TRUE always
 **/
#if 0
int geode_setup_gamma_table(GeodeDevRec *fPtr,LOCO *colour)
{
	int i,saturated=0;
	unsigned long rgb;

	for (i=0;i<256;i++) {
		WRITE_REG32(fPtr->pci_io,CS5530_PALETTE_ADDRESS,i);	/* palette index */
		if (colour[i].red > 255) {
			colour[i].red = 255;
			saturated++;
		}
		if (colour[i].green > 255) {
			colour[i].green = 255;
			saturated++;
		}
		if (colour[i].blue > 255) {
			colour[i].blue = 255;
			saturated++;
		}
		/*
		 * TODO check correct order
		 * rgb or bgr?
		 */
		rgb = colour[i].red;
		rgb <<= 8;
		rgb |= colour[i].green;
		rgb <<= 8;
		rgb |= colour[i].blue;
		WRITE_REG32(fPtr->pci_io,CS5530_PALETTE_DATA,rgb);	/* palette value */
	}
	/*
	 * video should bypass the palette, grafik should use the palette
	 */
	WRITE_REG32(fPtr->pci_io,CS5530_DISPLAY_CONFIG,
		    READ_REG32(fPtr->pci_io,CS5530_DISPLAY_CONFIG) & ~CS5530_DCFG_GV_PAL_BYP);

	if (saturated)
		xf86Msg(X_WARNING, "Gamma Setup: %d values saturated!\n",saturated);

	return TRUE;
}
#endif

/**
 * @brief enable compression
 * @param[in] fPtr hardware to work on
 **/
int geode_enable_compression(GeodeDevRec *fPtr)
{
	unsigned long tmp,cfg;

	/* disable current compression first */
	cfg=READ_REG32(fPtr->dc_io,DC_GENERAL_CFG) & ~(DC_GCFG_FDTY | DC_GCFG_CIM_MASK | DC_GCFG_CMPE | DC_GCFG_DECE);
	WRITE_REG32(fPtr->dc_io,DC_GENERAL_CFG,cfg);

	/* setup new compression data */
	WRITE_REG32(fPtr->dc_io,DC_CB_ST_OFFSET,fPtr->cbuffer_offset);
	tmp = READ_REG32(fPtr->dc_io,DC_LINE_DELTA) & DC_LINE_DELTA_FB_MASK;
	tmp |= (fPtr->cbuffer_ldelta>>2) << 12;
	WRITE_REG32(fPtr->dc_io,DC_LINE_DELTA,tmp);
	tmp =  READ_REG32(fPtr->dc_io,DC_BUF_SIZE) & (DC_BUF_SIZE_FB_MASK | DC_BUF_SIZE_VD_MASK);
	tmp |= ((fPtr->cbuffer_lsize>>2)+1) << 9;	/* see manual why '+1' */
	WRITE_REG32(fPtr->dc_io,DC_BUF_SIZE,tmp);

	/* enable the compression */
	WRITE_REG32(fPtr->dc_io,DC_GENERAL_CFG,cfg | DC_GCFG_CMPE | DC_GCFG_DECE);
	/* to be discussed: Makes it sense to insert one or more static
	 * frames into screen refresh? Does it hurt the user? I can see
	 * some artefacts only on fast moving content (Gaming).
	 */
	return TRUE;
}

/**
 * @brief setup the colour palette
 * @param[in] fPtr
 * @param[in] numColours
 * @param[in] indices
 * @param[in] colours
 * @param[in] pVisual FIXME
 *
 * Native routine to load the colour tabel into the DACs
 **/
int geode_load_colour_palette(GeodeDevRec *fPtr, int numColours, int *indices, LOCO *colours, VisualRec *pVisual)
{
	return FALSE;
}

/* -------------- a few internal functions ---------------- */

/**
 * @brief expand an 8 bit value to 16 bit
 **/
static __inline__ unsigned long geode_expand_8_2_16(int val)
{
	unsigned long rc;

	rc=(unsigned long)val & 0xFF;
	return (rc << 8) | rc;
}

/**
 * @brief expand a 16 bit value to 32 bit
 **/
static __inline__ unsigned long geode_expand_16_2_32(int val)
{
	unsigned long rc;

	rc=(unsigned long)val & 0xFFFF;
	return (rc << 16) | rc;
}

static unsigned long read_bb0_base(void)
{
	unsigned long rc;

	__asm__ __volatile__ ( "pushl %ebx" );
	__asm__ __volatile__ ( "movl $0xFFFFFF0C, %ebx");
	__asm__ __volatile__ ( ".byte 0x0f" );	/* CPU_READ */
	__asm__ __volatile__ ( ".byte 0x3d" : : : "eax");
	__asm__ __volatile__ ( "popl %ebx" );
	__asm__ __volatile__ ( "movl %%eax, %0" : "=a" (rc):);

	return rc;
}

static unsigned long read_bb1_base(void)
{
	unsigned long rc;

	__asm__ __volatile__ ( "pushl %ebx" );
	__asm__ __volatile__ ( "movl $0xFFFFFF1C, %ebx");
	__asm__ __volatile__ ( ".byte 0x0f" );	/* CPU_READ */
	__asm__ __volatile__ ( ".byte 0x3d" : : : "eax");
	__asm__ __volatile__ ( "popl %ebx" );
	__asm__ __volatile__ ( "movl %%eax, %0" : "=a" (rc) :);

	return rc;
}

static unsigned long read_bb0_pointer(void)
{
	unsigned long rc;

	__asm__ __volatile__ ( "pushl %ebx" );
	__asm__ __volatile__ ( "movl $0xFFFFFF2C, %ebx");
	__asm__ __volatile__ ( ".byte 0x0f" );	/* CPU_READ */
	__asm__ __volatile__ ( ".byte 0x3d" : : : "eax");
	__asm__ __volatile__ ( "popl %ebx" );
	__asm__ __volatile__ ( "movl %%eax, %0" : "=a" (rc) :);

	return rc;
}
static unsigned long read_bb1_pointer(void)
{
	unsigned long rc;

	__asm__ __volatile__ ( "pushl %ebx" );
	__asm__ __volatile__ ( "movl $0xFFFFFF3C, %ebx");
	__asm__ __volatile__ ( ".byte 0x0f" );	/* CPU_READ */
	__asm__ __volatile__ ( ".byte 0x3d" : : : "eax");
	__asm__ __volatile__ ( "popl %ebx" );
	__asm__ __volatile__ ( "movl %%eax, %0" : "=a" (rc) :);

	return rc;
}

/**
 * @brief set a new base to the blit base 0
 * @param[in] value new base as offset in bytes
 *
 * Note: These instructions only work if the scratch pad SRAM
 * is *enabled*. Do not call this routine with scratch pad ram
 * disabled (means acceleration disabled). Then these instructions
 * fail, with "illegal instruction" (TBC)
 **/
static __inline__ void set_bb0_base(unsigned long value)
{
	__asm__ __volatile__ ( "pushl %ebx" );
	__asm__ __volatile__ ( "movl $0xFFFFFF0C, %ebx");
	__asm__ __volatile__ ( "movl %0, %%eax": : "r" (value) );

	__asm__ __volatile__ ( ".byte 0x0f" );	/* CPU_WRITE */
	__asm__ __volatile__ ( ".byte 0x3c" );

	__asm__ __volatile__ ( ".byte 0x0f" );	/* BB0_RESET */
	__asm__ __volatile__ ( ".byte 0x3a" );
	__asm__ __volatile__ ( "popl %ebx" );
}

/**
 * @brief set a new base to the blit base 1
 * @param[in] value new base as offset in bytes
 *
 * Note: These instructions only work if the scratch pad SRAM
 * is *enabled*. Do not call this routine with scratch pad ram
 * disabled (means acceleration disabled). Then these instructions
 * fail, with "illegal instruction" (TBC)
 */
static __inline__ void set_bb1_base(unsigned long value)
{
	__asm__ __volatile__ ( "pushl %ebx" );
	__asm__ __volatile__ ( "movl $0xFFFFFF1C, %ebx");
	__asm__ __volatile__ ( "movl %0, %%eax": : "r" (value) );

	__asm__ __volatile__ ( ".byte 0x0f" );	/* CPU_WRITE */
	__asm__ __volatile__ ( ".byte 0x3c" );

	__asm__ __volatile__ ( ".byte 0x0f" );	/* BB1_RESET */
	__asm__ __volatile__ ( ".byte 0x3b" );
	__asm__ __volatile__ ( "popl %ebx" );
}

/**
 * @brief wait until a blit is pending in the chipset
 * @param[in] fPtr Pointer to manage the hardware
 *
 * This waits for new register loading
 **/
static void geode_wait_until_blit_pending(GeodeDevRec *fPtr)
{
	/* FIXME use an timeout! */
	while(READ_REG16(fPtr->gp_io,GP_BLIT_STATUS) & BS_BLIT_PENDING) { ; }
}

/**
 * @brief wait until last blit is finished
 * @param[in] fPtr Pointer to manage the hardware
 *
 * This waits until video memory is free to access
 **/
static void geode_wait_until_busy(GeodeDevRec *fPtr)
{
	/* FIXME use an timeout! */
	while(READ_REG16(fPtr->gp_io,GP_BLIT_STATUS) & BS_BLIT_BUSY) { ; }
}

/**
 * @brief setup the blit buffer pointers for acceleration
 * @param[in] fPtr FIXME
 * @param[in] ram_size size in bytes the scratch pad ram contains
 * @param[in] ram_offset first active cell offset from base mapping
 *
 * Some notes about the scratch pad RAM:
 * First of all: Its a very ugly design!. It uses a part of the
 * already small CPU cache and needs many preparations to get it work.
 * Setup of this RAM is done in kernel's framebuffer driver
 * (cache line mapping, it requires to disable/reenable the cache).
 *
 * We do the rest of the setup here.
 *
 * I/O space memory mapping of this area:
 *
 * ###############################################
 * ^_base address                                ^_+4096
 *
 * Possible configurations:
 *                      ########################## 2048 bytes size
 *                      ^_bb0       ^_bb1
 *        ######################################## 3072 bytes size
 *        ^_bb0               ^_bb1
 * ############################################### 4096 bytes size
 * ^_bb0                ^_bb1
 *
 * On a standard VM BIOS this does not work, as the SMM handler
 * uses space in this memory. But we do not use SMM here, so
 * we can use the whole area (but we are also responsible for
 * the correct setup...)!
 *
 * @return TRUE (always)
 **/
int geode_setup_blitbuffer_pointers(GeodeDevRec * fPtr,int ram_size, int ram_offset)
{
	set_bb0_base(0);		/* always at the base */
	set_bb1_base(ram_size/2);	/* always in the middle at 16 byte boundary */
	/*
	 * calculate both offsets in the scratch pad ram
	 * fb driver will provide us with the startoffset and always
	 * sets BB0 to the start, BB1 to the half of the size of the scratch pad ram
	 */
	fPtr->Geodebb0Base = ram_offset;
	fPtr->Geodebb1Base = ram_offset + ram_size/2;

	fPtr->GeodebufferWidthPixels = fPtr->Geodebb1Base - fPtr->Geodebb0Base;

	return TRUE;
}


/**
 * @brief This routine is used to specify a solid pattern colour
 * @param[in] fPtr Pointer to manage the hardware
 * @param[in] colour Colour to draw
 *
 * It is called before performing solid rectangle fills or more
 * complicated BLTs that use a solid pattern colour.
 **/
static void geode_set_solid_pattern(GeodeDevRec *fPtr,int colour)
{
	fPtr->GFXsourceFlags = 0;	/* CLEAR TRANSPARENCY FLAG */
	fPtr->GFXpatternFlags = 0;	/* SET PATTERN FLAGS */

	if (fPtr->geode_not_native)
		fPtr->GFXsavedColour = geode_expand_8_2_16(colour);
	else
		fPtr->GFXsavedColour = colour;

	/* POLL UNTIL ABLE TO WRITE THE PATTERN COLOR */
	geode_wait_until_blit_pending(fPtr);

	WRITE_REG16(fPtr->gp_io,GP_PAT_COLOR_0, fPtr->GFXsavedColour);
}

/**
 * @brief loads the specified raster operation
 * @param[in] fPtr Pointer to manage the hardware
 * @param[in] rop real raster operation to do
 *
 * It sets the pattern flags appropriately
 **/
static void geode_set_raster_operation(GeodeDevRec *fPtr,int rop)
{
	unsigned short rop16;

	/* GENERATE 16-BIT VERSION OF ROP WITH PATTERN FLAGS */

	rop16 = (unsigned short)rop | fPtr->GFXpatternFlags;
	if ((rop & 0x33) ^ ((rop >> 2) & 0x33))
		rop16 |= fPtr->GFXsourceFlags;

	/* SAVE ROP FOR LATER COMPARISONS */
	/* Need to have the pattern flags included */

	fPtr->GFXsavedRop = rop16;

	/* SET FLAG INDICATING ROP REQUIRES DESTINATION DATA */
	/* True if even bits (0:2:4:6) do not equal the correspinding */
	/* even bits (1:3:5:7). */

	fPtr->GFXusesDstData = ((rop & 0x55) ^ ((rop >> 1) & 0x55));

	/* POLL UNTIL ABLE TO WRITE THE PATTERN COLOR */
	/* Only one operation can be pending at a time. */

	geode_wait_until_blit_pending(fPtr);
	WRITE_REG16(fPtr->gp_io,GP_RASTER_MODE,rop16);
}

/**
 * @brief used to specify a solid source colour
 * @param[in] fPtr Pointer to manage the hardware
 * @param[in] source pattern
 **/
static void geode_set_solid_source(GeodeDevRec *fPtr,int source)
{
	/* CLEAR TRANSPARENCY FLAG */

	fPtr->GFXsourceFlags = 0;

	/* FORMAT 8 BPP COLOR */
	/* GX requires 8BPP colour data be duplicated into bits [15:8]. */

	if (fPtr->geode_not_native)
		source = geode_expand_8_2_16(source);

	/* POLL UNTIL ABLE TO WRITE THE SOURCE COLOR */

	geode_wait_until_blit_pending(fPtr);
	WRITE_REG16(fPtr->gp_io,GP_SRC_COLOR_0, source);
	WRITE_REG16(fPtr->gp_io,GP_SRC_COLOR_1, source);
}

/**
 * @brief specify a monochrome pattern and its colours
 * @param[in] fPtr Pointer to manage the hardware
 * @param[in] bgcolour FIXME
 * @param[in] fgcolour FIXME
 * @param[in] data0 FIXME
 * @param[in] data1 FIXME
 * @param[in] transparent true if required
 **/
static void geode_set_mono_pattern(GeodeDevRec *fPtr, unsigned long bgcolour, unsigned long fgcolour,
			unsigned long data0, unsigned long data1, int transparent)
{
	/* CLEAR TRANSPARENCY FLAG */

	fPtr->GFXsourceFlags = 0;

	/* SET PATTERN FLAGS */

	fPtr->GFXpatternFlags = transparent ? RM_PAT_MONO | RM_PAT_TRANSPARENT : RM_PAT_MONO;

	if (fPtr->geode_not_native) {
		bgcolour = geode_expand_8_2_16(bgcolour);
		fgcolour = geode_expand_8_2_16(fgcolour);
	}

	/* POLL UNTIL ABLE TO WRITE THE PATTERN COLORS AND DATA */

	geode_wait_until_blit_pending(fPtr);;
	WRITE_REG16(fPtr->gp_io,GP_PAT_COLOR_0, bgcolour);
	WRITE_REG16(fPtr->gp_io,GP_PAT_COLOR_1, fgcolour);
	WRITE_REG32(fPtr->gp_io,GP_PAT_DATA_0, data0);
	WRITE_REG32(fPtr->gp_io,GP_PAT_DATA_1, data1);
}

/**
 * @brief specify the monochrome source colours
 * @param[in] fPtr Pointer to manage the hardware
 * @param[in] bgcolour FIXME
 * @param[in] fgcolour FIXME
 * @param[in] transparent FIXME
 *
 * used to specify the monochrome source colours. It must be called
 * *after* loading any pattern data (those routines clear the source flags)
 **/
static void geode_set_mono_source(GeodeDevRec *fPtr, unsigned long bgcolour,
			unsigned long fgcolour, unsigned short transparent)
{
	/* SET TRANSPARENCY FLAG */

	fPtr->GFXsourceFlags = transparent ? RM_SRC_TRANSPARENT : 0;

	/* FORMAT 8 BPP COLOR */
	/* GX requires 8BPP colour data be duplicated into bits [15:8]. */

	if (fPtr->geode_not_native) {
		bgcolour = geode_expand_8_2_16(bgcolour);
		fgcolour = geode_expand_8_2_16(fgcolour);
	}

	/* POLL UNTIL ABLE TO WRITE THE SOURCE COLOR */

	geode_wait_until_blit_pending(fPtr);;
	WRITE_REG16(fPtr->gp_io,GP_SRC_COLOR_0, bgcolour);
	WRITE_REG16(fPtr->gp_io,GP_SRC_COLOR_1, fgcolour);
}

/**
 * @brief MUST be used when performing a solid rectangle fill
 * @param[in] fPtr Pointer to manage the hardware
 * @param[in] x screen X position (left)
 * @param[in] y screen Y position (top)
 * @param[in] width width of rectangle, in pixels
 * @param[in] height height of rectangle, in scanlines
 * @param[in] colour fill colour
 *
 * This routine MUST be used when performing a solid rectangle fill with
 * the ROPs of PATCOPY (0xF0), BLACKNESS (0x00), WHITENESS (0xFF), or
 * PATINVERT (0x0F). There is a bug in GXm for these cases that requires a
 * workaround.
 *
 * For BLACKNESS (ROP = 0x00), set the colour to 0x0000.
 * For WHITENESS (ROP = 0xFF), set the colour to 0xFFFF.
 * For PATINVERT (ROP = 0x0F), invert the desired colour.
 **/
static void geode_solid_fill(GeodeDevRec *fPtr,unsigned short x, unsigned short y,unsigned short width,
			unsigned short height, int colour)
{
	/* unsigned short section; */

	/* POLL UNTIL ABLE TO WRITE TO THE REGISTERS */
	/* Only one operation can be pending at a time. */

	geode_wait_until_blit_pending(fPtr);

	/* SET REGISTERS TO DRAW RECTANGLE */

	WRITE_REG16(fPtr->gp_io,GP_DST_XCOOR, x);
	WRITE_REG16(fPtr->gp_io,GP_DST_YCOOR, y);
	WRITE_REG16(fPtr->gp_io,GP_HEIGHT, height);
	WRITE_REG16(fPtr->gp_io,GP_RASTER_MODE, 0x00F0);	/* PATCOPY */
	WRITE_REG16(fPtr->gp_io,GP_PAT_COLOR_0, colour);
	WRITE_REG16(fPtr->gp_io,GP_WIDTH, width);
	WRITE_REG16(fPtr->gp_io,GP_BLIT_MODE, 0);
}


/**
 * @brief render a rectangle using the current raster operation and the specified colour pattern
 * @param[in] fPtr FIXME
 * @param[in] x screen X position (left)
 * @param[in] y screen Y position (top)
 * @param[in] width width of rectangle, in pixels
 * @param[in] height height of rectangle, in scanlines
 * @param[in] pattern pointer to 8x8 colour pattern data
 *
 * It allows an 8x8 colour pattern to be rendered without multiple
 * calls to the set_colour_pattern and pattern_fill routines
 **/
static void geode_colour_pattern_fill(GeodeDevRec *fPtr,unsigned short x, unsigned short y,
		       unsigned short width, unsigned short height, unsigned long *pattern)
{
	unsigned short blit_mode, passes, cur_y, pat_y, i;
	unsigned short buffer_width, line_width;
	unsigned short bpp_shift, section, cur_x;

	/* SET APPROPRIATE INCREMENT */

	bpp_shift = (!fPtr->geode_not_native) ? 2 : 1;

	/* SET DESTINATION REQUIRED */

	blit_mode = fPtr->GFXusesDstData ? BM_READ_DST_FB0 : 0;

	/* SET SOURCE EXPANSION */

	blit_mode |= BM_SOURCE_EXPAND;

	/* OVERRIDE RASTER MODE TO FORCE A COLOR PATTERN */

	geode_wait_until_blit_pending(fPtr);
	WRITE_REG16(fPtr->gp_io,GP_RASTER_MODE,
		    (fPtr->GFXsavedRop & ~RM_PAT_MASK & ~RM_PAT_TRANSPARENT) |
				    RM_PAT_COLOR);

	/* WRITE THE REGISTERS THAT DO NOT CHANGE         */
	/* If destination data is required, the width and */
	/* x position will be overwritten.                */

	WRITE_REG16(fPtr->gp_io,GP_HEIGHT, 1);
	WRITE_REG16(fPtr->gp_io,GP_WIDTH, width);
	WRITE_REG16(fPtr->gp_io,GP_DST_XCOOR, x);

	/* THE ENTIRE PATTERN WILL NOT BE DRAWN IF THE HEIGHT IS LESS THAN 8 */

	passes = (height < 8) ? height : 8;

	/* SINCE ONLY DESTINATION DATA, WE CAN USE BOTH BB0 AND BB1. */
	/* Therefore, width available = BLT buffer width * 2. */

	buffer_width = fPtr->GeodebufferWidthPixels << 1;

	for (i = 0; i < passes; i++) {
		pat_y = ((y + i) & 7) << bpp_shift;
		cur_y = y + i;

		/* WRITE THE PATTERN DATA FOR THE ACTIVE LINE */

		geode_wait_until_blit_pending(fPtr);
		WRITE_REG32(fPtr->gp_io,GP_PAT_DATA_0, pattern[pat_y]);
		WRITE_REG32(fPtr->gp_io,GP_PAT_DATA_1, pattern[pat_y + 1]);

		if (!fPtr->geode_not_native) {
			WRITE_REG32(fPtr->gp_io,GP_PAT_DATA_2, pattern[pat_y + 2]);
			WRITE_REG32(fPtr->gp_io,GP_PAT_DATA_3, pattern[pat_y + 3]);
		}

		/* SPLIT BLT LINE INTO SECTIONS IF REQUIRED              */
		/* If no destination data is required, we can ignore     */
		/* the BLT buffers.  Otherwise, we must separate the BLT */
		/* so as not to overflow the buffers                     */

		if (blit_mode & BM_READ_DST_BB0) {
			line_width = width;
			cur_x = x;

			while (line_width) {
				section = (line_width > buffer_width) ? buffer_width : line_width;
				cur_y = y + i;

				geode_wait_until_blit_pending(fPtr);
				WRITE_REG16(fPtr->gp_io,GP_DST_XCOOR, cur_x);
				WRITE_REG16(fPtr->gp_io,GP_WIDTH, section);

				while (cur_y < y + height) {
					geode_wait_until_blit_pending(fPtr);
					WRITE_REG16(fPtr->gp_io,GP_DST_YCOOR, cur_y);
					WRITE_REG16(fPtr->gp_io,GP_BLIT_MODE, blit_mode);
					cur_y += 8;
				}

				cur_x += section;
				line_width -= section;
			}

		} else {
			while (cur_y < y + height) {
				geode_wait_until_blit_pending(fPtr);
				WRITE_REG16(fPtr->gp_io,GP_DST_YCOOR, cur_y);
				WRITE_REG16(fPtr->gp_io,GP_BLIT_MODE, blit_mode);
				cur_y += 8;
			}
		}
	}

	/* RESTORE ORIGINAL ROP AND FLAGS */

	geode_wait_until_blit_pending(fPtr);
	WRITE_REG16(fPtr->gp_io,GP_RASTER_MODE, fPtr->GFXsavedRop);
}


/* ################### X Interface Routines ######################### */

/* ------------------------------------------------------------------ */

/**
 * @brief set up the planemask and raster for future bliting functionality
 * @param[in] pScreenInfo FIXME
 * @param[in] xdir horizontal direction (-1 or +1) to copy
 * @param[in] ydir vertical direction (-1 or +1) to copy
 * @param[in] X11rop one of X11's raster operation
 * @param[in] planemask value of masking from rop data
 * @param[in] transparency_colour colour of pixels not to copy from src to dst
 *
 * The patterns specified is ignored inside the function
 **/
static void geode_setup_for_screen_to_screen_copy(ScrnInfoRec *pScreenInfo,int xdir, int ydir, int rop,
					unsigned int planemask, int transparency_colour)
{
	int components;
	GeodeDevRec *fPtr = GEODEDEVPTR( pScreenInfo );

	/* xf86Msg(X_WARNING, "%s %d/%d:%d->%04lX:%08X,%d\n",__FUNCTION__,xdir,ydir,rop,rop16,planemask,transparency_colour); */

	fPtr->GFXsavedRop = XAAGetCopyROP(rop);

	/* SET FLAG INDICATING ROP REQUIRES DESTINATION DATA */
	if (planemask != 0xFFFFFFFF) {
		xf86Msg(X_WARNING,"Screen copy wit planemask != -1 (=0x%08X)!\n",planemask);
		components = copy_components_pm[rop];
	} else
		components = copy_components[rop];

	/* FORMAT 8 BPP COLOR */
	/* GX requires 8BPP colour data be duplicated into bits [15:8]. */

	if (fPtr->geode_not_native)
		planemask = geode_expand_8_2_16(planemask);

	fPtr->Geode_blt_mode = BM_READ_SRC_NONE | BM_READ_DST_NONE;
	/*
	 * decide and set if source and/or destination data is required
	 */
	if (components & ROP_SRC) {
		fPtr->Geode_blt_mode |= BM_READ_SRC_FB;	/* this uses BLT buffer 0 */
		if (components & ROP_DST) {
			/*
			 * we need the destination data also. So we must
			 * half the BLT buffer
			 */
			fPtr->Geode_blt_mode |= BM_READ_DST_FB1;	/* dest data into BLT buffer 1 */
			fPtr->Geode_buffer_width = fPtr->GeodebufferWidthPixels;	/* half size */
			/* xf86Msg(X_WARNING,"Screen copy with source and dest rop=%d (%u)!\n",rop,fPtr->Geode_buffer_width); */
		}
		else {
			/* source data only. We can use the full BLT buffer */
			fPtr->Geode_buffer_width = fPtr->GeodebufferWidthPixels << 1;
			/* xf86Msg(X_WARNING,"Screen copy with source only rop=%d (%u)!\n",rop,fPtr->Geode_buffer_width); */
		}
	} else {
		if (components & ROP_DST) {
			/* only destination data is required */
			fPtr->Geode_blt_mode |= BM_READ_DST_FB0;
			fPtr->Geode_buffer_width = fPtr->GeodebufferWidthPixels << 1;
			/* xf86Msg(X_WARNING,"Screen copy with dest only rop=%d (%u)!\n",rop,fPtr->Geode_buffer_width); */
		} else {
			/* wheter DST nor SRC data is required. Seems a solid ROP */
			/* Can use both BLT buffers if no destination data is required. */
			fPtr->Geode_buffer_width = fPtr->GeodebufferWidthPixels << 1;
			xf86Msg(X_WARNING,"Screen copy without source and dest rop=%d (%u)!\n",rop,fPtr->Geode_buffer_width);
		}
	}

	/* POLL UNTIL ABLE TO WRITE THE PATTERN COLOR */
	/* Only one operation can be pending at a time. */

	geode_wait_until_blit_pending(fPtr);

	WRITE_REG16(fPtr->gp_io,GP_PAT_COLOR_0, (unsigned short)planemask);
	WRITE_REG16(fPtr->gp_io,GP_RASTER_MODE, fPtr->GFXsavedRop);

	fPtr->GeodeTransparent = (transparency_colour == -1) ? 0 : 1;
	fPtr->GeodeTransColour = transparency_colour;
}

/**
 * @brief perform a screen to screen BLT
 * @param[in] pScreenInfo FIXME
 * @param[in] srcx x-coordinate of the source window (upper left hand corner)
 * @param[in] srcy y-coordinate of the source window (upper left hand corner)
 * @param[in] dstx x-coordinate of the destination window (upper left hand corner)
 * @param[in] dsty y-coordinate of the destination window (upper left hand corner)
 * @param[in] width of the window to be copied
 * @param[in] height of the window to be copied
 *
 * BLT is done on previously specified planemask, raster
 * operation and transparency flag
 *
 * The patterns specified is ignored inside the function
 **/
static void geode_subsequent_screen_to_screen_copy(ScrnInfoRec *pScreenInfo,
		int srcx, int srcy, int dstx, int dsty, int width, int height)
{
	int section;
	unsigned short blit_mode = 0;
	GeodeDevRec *fPtr = GEODEDEVPTR( pScreenInfo );

	/* xf86Msg(X_WARNING, "%s %d/%d/%d/%d -> %d/%d\n",__FUNCTION__,srcx,srcy,width,height,dstx,dsty); */

	if (fPtr->GeodeTransparent) {
		if (fPtr->geode_not_native)
			fPtr->GeodeTransColour = geode_expand_8_2_16(fPtr->GeodeTransColour);
		fPtr->GeodeTransColour = geode_expand_16_2_32(fPtr->GeodeTransColour);

		/* WAIT UNTIL PIPELINE IS NOT BUSY BEFORE LOADING DATA INTO BB1 */
		/* Need to make sure any previous BLT using BB1 is complete. */
		/* Only need to load 32 bits of BB1 for the 1 pixel BLT that follows. */

		geode_wait_until_busy(fPtr);
		WRITE_REG32(fPtr->scratch_io,fPtr->Geodebb1Base,fPtr->GeodeTransColour);

		/* DO BOGUS BLT TO LATCH DATA FROM BB1 */
		/* Already know graphics pipeline is idle. */
		/* Only need to latch data into the holding registers for the current */
		/* data from BB1.  A 1 pixel wide BLT will suffice. */

		WRITE_REG32(fPtr->gp_io,GP_DST_XCOOR, 0);
		WRITE_REG32(fPtr->gp_io,GP_SRC_XCOOR, 0);
		WRITE_REG32(fPtr->gp_io,GP_WIDTH, 0x00010001);
		WRITE_REG16(fPtr->gp_io,GP_RASTER_MODE, 0x00CC);
		WRITE_REG16(fPtr->gp_io,GP_BLIT_MODE, BM_READ_SRC_FB | BM_READ_DST_BB1);

		/* WRITE REGISTERS FOR REAL SCREEN TO SCREEN BLT */

		geode_wait_until_blit_pending(fPtr);

		WRITE_REG16(fPtr->gp_io,GP_HEIGHT, height);	/* FIXME why? */
		WRITE_REG16(fPtr->gp_io,GP_RASTER_MODE, 0x10C6);	/* FIXME what is this mode */
		WRITE_REG32(fPtr->gp_io,GP_PAT_COLOR_0, 0xFFFFFFFF);	/* FIXME why a new value? */
	}

	/* CHECK Y DIRECTION */
	/* Hardware has support for negative Y direction. */

	if (dsty > srcy) {
		blit_mode = BM_REVERSE_Y;
		srcy += height - 1;
		dsty += height - 1;
	}

	/* CHECK X DIRECTION */
	/* Hardware does not support negative X direction since at the time */
	/* of development all supported resolutions could fit a scanline of */
	/* data at once into the BLT buffers (using both BB0 and BB1).  This */
	/* code is more generic to allow for any size BLT buffer. */

	if (dstx > srcx) {
		srcx += width;
		dstx += width;
	}

	/* POLL UNTIL ABLE TO WRITE TO THE REGISTERS */
	/* Write the registers that do not change for each section. */

	/* xf86Msg(X_WARNING, "%s ",__FUNCTION__); */
	geode_wait_until_blit_pending(fPtr);
	WRITE_REG16(fPtr->gp_io,GP_HEIGHT, height);

	/* REPEAT UNTIL FINISHED WITH RECTANGLE */
	/* Perform BLT in vertical sections, as wide as the BLT buffer allows. */
	/* Hardware does not split the operations, so software must do it to */
	/* avoid large scanlines that would overflow the BLT buffers. */

	while (width > 0) {
		/* CHECK WIDTH OF CURRENT SECTION */

		if (width > fPtr->Geode_buffer_width)
			section = fPtr->Geode_buffer_width;
		else
			section = width;

		/* PROGRAM REGISTERS THAT ARE THE SAME FOR EITHER X DIRECTION */

		geode_wait_until_blit_pending(fPtr);
		WRITE_REG16(fPtr->gp_io,GP_SRC_YCOOR, srcy);
		WRITE_REG16(fPtr->gp_io,GP_DST_YCOOR, dsty);
		WRITE_REG16(fPtr->gp_io,GP_WIDTH, section);

		/* CHECK X DIRECTION */

		if (dstx > srcx) {
			/* NEGATIVE X DIRECTION */
			/* Still positive X direction within the section. */

			srcx -= section;
			dstx -= section;
			WRITE_REG16(fPtr->gp_io,GP_SRC_XCOOR, srcx);
			WRITE_REG16(fPtr->gp_io,GP_DST_XCOOR, dstx);
		} else {
			/* POSITIVE X DIRECTION */

			WRITE_REG16(fPtr->gp_io,GP_SRC_XCOOR, srcx);
			WRITE_REG16(fPtr->gp_io,GP_DST_XCOOR, dstx);
			dstx += section;
			srcx += section;
		}
		/* xf86Msg(X_WARNING, "(%d/%d|%d/%d -> %d/%d)",srcx,srcy,width,height,dstx,dsty); */
		WRITE_REG16(fPtr->gp_io,GP_BLIT_MODE, fPtr->Geode_blt_mode | blit_mode);
		width -= section;
	}
	/* xf86Msg(X_WARNING, "\n"); */
}

/* ------------ 8x8 colour pattern ------------------------------ */

/**
 * @brief fill the colour pattern of 8x8
 * @param[in] pScreenInfo FIXME
 * @param[in] patterx set from on rop data
 * @param[in] patterny set based on rop data
 * @param[in] planemask value of masking from rop data
 * @param[in] trans_colour FIXME
 **/
static void geode_setup_for_8x8_pattern_colour_expand(ScrnInfoPtr pScreenInfo,
		int patternx, int patterny,int rop, unsigned int planemask,int trans_colour)
{
	GeodeDevRec *fPtr = GEODEDEVPTR( pScreenInfo );

	/* CHECK IF PLANEMASK IS NOT USED (ALL PLANES ENABLED) */
	if (planemask == 0xFFFFFFFF) {
		/* USE NORMAL PATTERN ROPs IF ALL PLANES ARE ENABLED */
		geode_set_raster_operation(fPtr,XAAGetPatternROP(rop));
	} else {
		/* SELECT ROP THAT USES SOURCE DATA FOR PLANEMASK */
		geode_set_solid_source(fPtr,planemask);
		geode_set_raster_operation(fPtr,XAAGetPatternROP_PM(rop));
	}
}

/**
 * @brief called to fill  a rectangle with the colour pattern of previously loaded pattern
 * @param[in] pScreenInfo FIXME
 * @param[in] patterx set from on rop data
 * @param[in] pattery set based on rop data
 * @param[in] x x-coordinates of the destination rectangle
 * @param[in] y y-coordinates of the destination rectangle
 * @param[in] w Specifies width of the rectangle
 * @param[in] h Height of the window of the rectangle
 **/
static void geode_subsequent_8x8_pattern_colour_expand(ScrnInfoPtr pScreenInfo,
		int patternx, int patterny, int x, int y,int w, int h)
{
	GeodeDevRec *fPtr = GEODEDEVPTR( pScreenInfo );

	/* Ignores specified pattern. */
	/* FIXME: Who set the pattern into the memory */
	geode_colour_pattern_fill(fPtr,(unsigned short)x, (unsigned short)y,
	    (unsigned short)w, (unsigned short)h,
	    ((unsigned long *)((fPtr->fbstart +
			(patterny << fPtr->yshift)) + patternx)));
}

/* ---------------- filling solid pattern --------------------------- */

/**
 * @brief This routine is called to setup the solid pattern
 *        colour for future rectangular fills or vectors
 * @param[in] pScreenInfo FIXME
 * @param[in] colour Specifies the colour to be filled up in defined area
 * @param[in] rop Specifies the X raster operation value
 * @param[in] planemask Specifies the masking value based rop srcdata
 **/
static void geode_setup_for_fill_rect(ScrnInfoRec *pScreenInfo, int colour, int rop, unsigned int planemask)
{
	GeodeDevRec *fPtr = GEODEDEVPTR( pScreenInfo );
	geode_set_solid_pattern(fPtr, colour&0xFFFF);

	/* CHECK IF PLANEMASK IS NOT USED (ALL PLANES ENABLED) */
	if (planemask == 0xFFFFFFFF) {
		/* USE NORMAL PATTERN ROPs IF ALL PLANES ARE ENABLED */
/*		xf86Msg(X_WARNING, "Setup for normal pattern.\n"); */
		/* xf86Msg(X_INFO, "%s: (%X,%d=%X)\n",__FUNCTION__,colour&0xFFFF,rop,XAAGetPatternROP(rop)); */
		geode_set_raster_operation(fPtr,XAAGetPatternROP(rop));
	} else {
		/* SELECT ROP THAT USES SOURCE DATA FOR PLANEMASK */
/*		xf86Msg(X_WARNING, "Setup for source data pattern.\n"); */
		/*
		 * For the Xfree86 display driver, the source colour is used to
		 * specify a planemask and the ROP is adjusted accordingly.
		 */
		xf86Msg(X_INFO, "%s:pl (%X,%d=%X)\n",__FUNCTION__,colour,rop,XAAGetPatternROP_PM(rop));
		geode_set_solid_source(fPtr,planemask);
		geode_set_raster_operation(fPtr,XAAGetPatternROP_PM(rop));
	}
}

/**
 * @brief render a rectangle using the current raster operation and the specified colour pattern
 * @param[in] fPtr Pointer to manage the hardware
 * @param[in] x x co-ordinates (left)
 * @param[in] y y co-ordinates (top)
 * @param[in] width width to fill
 * @param[in] height height respectively
 *
 * This function still leave visible garbage on screen. Some blits seems not to work!
 * It works, but only *without* the workaround (16 bit alignement)!
 *
 * Note: This function will also be used by 8x8 monocrome fill
 **/
static void geode_subsequent_fill_rect_solid(ScrnInfoRec *pScreenInfo,int x, int y, int width, int height)
{
	GeodeDevRec *fPtr = GEODEDEVPTR( pScreenInfo );
	int section, buffer_width, blit_mode;
	/* int fill; */

	/* xf86Msg(X_INFO, "%s: (%d,%d,%d,%d)\n",__FUNCTION__,x,y,width,height); */
	/* CHECK IF OPTIMIZED SOLID CASES */
	/* Check all 16 bits of the ROP to include solid pattern flags. */

	switch (fPtr->GFXsavedRop) {
		/* CHECK FOR SPECIAL CASES WITHOUT DESTINATION DATA */
		/* Need hardware workaround for fast "burst write" cases. */
	case 0x00F0: /* COPY */
		geode_solid_fill(fPtr, x, y, width, height, fPtr->GFXsavedColour);
		break;

	case 0x000F: /* INVERT */
		geode_solid_fill(fPtr, x, y, width, height, ~fPtr->GFXsavedColour);
		break;

	case 0x0000: /* Black */
		geode_solid_fill(fPtr, x, y, width, height, 0x0000);
		break;

	case 0x00FF: /* White */
		geode_solid_fill(fPtr, x, y, width, height, 0xFFFF);
		break;

	default:
		/* REMAINING CASES REQUIRE DESTINATION DATA OR NOT SOLID COLOR */
		/* DETERMINE BLT MODE VALUE */
		/* Still here for non-solid patterns without destination data. */
		blit_mode = fPtr->GFXusesDstData ? BM_READ_DST_FB0 : 0;

		/* SET SOURCE EXPANSION MODE */
		/* If the ROP requires source data, then the source data is all 1's */
		/* and then expanded into the desired colour in GP_SRC_COLOR_1.
		 * FIXME: GP_SRC_COLOR_1 is only set, when planemask is not -1
		 */
		blit_mode |= BM_SOURCE_EXPAND;

		/* POLL UNTIL ABLE TO WRITE TO THE REGISTERS */
		/* Write the registers that do not change for each section. */

		geode_wait_until_blit_pending(fPtr);
		WRITE_REG16(fPtr->gp_io,GP_HEIGHT, height);

		if (width <= 16) {
			/* OK TO DRAW SMALL RECTANGLE IN ONE PASS */
			WRITE_REG16(fPtr->gp_io,GP_DST_XCOOR, x);
			WRITE_REG16(fPtr->gp_io,GP_DST_YCOOR, y);
			WRITE_REG16(fPtr->gp_io,GP_WIDTH, width);
			WRITE_REG16(fPtr->gp_io,GP_BLIT_MODE, blit_mode);
		} else {
			/* SINCE ONLY DESTINATION DATA, WE CAN USE BOTH BB0 AND BB1. */
			/* Therefore, width available = BLT buffer width * 2. */

			buffer_width = fPtr->GeodebufferWidthPixels << 1;

			/* REPEAT UNTIL FINISHED WITH RECTANGLE */
			/* Perform BLT in vertical sections, as wide as the BLT buffer */
			/* allows.  Hardware does not split the operations, so */
			/* software must do it to avoid large scanlines that would */
			/* overflow the BLT buffers. */
			while (width > 0) {
				/* DETERMINE WIDTH OF SECTION */
				if (width > buffer_width)
					section = buffer_width;
				else
					section = width;
# if 0
				fill = 0x10 - (x & 0x0F);
				if (fill) {
					geode_wait_until_blit_pending(fPtr);
					WRITE_REG16(fPtr->gp_io,GP_DST_XCOOR, x);
					WRITE_REG16(fPtr->gp_io,GP_DST_YCOOR, y);
					WRITE_REG16(fPtr->gp_io,GP_WIDTH, fill);
					WRITE_REG16(fPtr->gp_io,GP_BLIT_MODE, blit_mode);
					width -= fill;
					section -= fill;
					x += fill;
				}
# endif
			/* POLL UNTIL ABLE TO WRITE TO THE REGISTERS */
			/* if (x & 0xF)
				xf86Msg(X_WARNING, "%s X pos starts at %d!\n",__FUNCTION__,x); */

			geode_wait_until_blit_pending(fPtr);

			WRITE_REG16(fPtr->gp_io,GP_DST_XCOOR, x);
			WRITE_REG16(fPtr->gp_io,GP_DST_YCOOR, y);
			WRITE_REG16(fPtr->gp_io,GP_WIDTH, section);
			WRITE_REG16(fPtr->gp_io,GP_BLIT_MODE, blit_mode);

			/* ADJUST PARAMETERS FOR NEXT SECTION */

			width -= section;
			x += section;
			}
		}
		break;
	}
}

/* ------------ 8x8 monochrome pattern ------------------------------ */

/**
 * @brief called to fill the monochrome pattern of 8x8
 * @param[in] pScreenInfo FIXME
 * @param[in] pattern0 first DWORD of pattern
 * @param[in] pattern1 second DWORD of pattern
 * @param[in] fg foreground colour for 1 bits in pattern
 * @param[in] bg background colour for 0 bits in pattern
 * @param[in] rop raster operation
 * @param[in] planemask value of masking from rop data
 *
 * Note: As we set the HARDWARE_PATTERN_SCREEN_ORIGIN and HARDWARE_PATTERN_PROGRAMMED_BITS
 * bits, we always get the pattern itself instead of a position.
 **/
static void geode_setup_for_8x8_pattern_mono_expand(ScrnInfoPtr pScreenInfo,
		int pattern0, int pattern1, int fg, int bg, int rop, unsigned int planemask)
{
	GeodeDevRec *fPtr = GEODEDEVPTR( pScreenInfo );
	int trans = (bg == -1);

	/* LOAD PATTERN COLORS AND DATA */
	geode_set_mono_pattern(fPtr, bg, fg, pattern0, pattern1, trans);
	geode_set_mono_source(fPtr, bg, fg, trans);

	/* CHECK IF PLANEMASK IS NOT USED (ALL PLANES ENABLED) */
	if (planemask == 0xFFFFFFFF) {
		/* USE NORMAL PATTERN ROPs IF ALL PLANES ARE ENABLED */
		geode_set_raster_operation(fPtr,XAAGetPatternROP(rop));
	} else {
		/* SELECT ROP THAT USES SOURCE DATA FOR PLANEMASK */
		geode_set_solid_source(fPtr,planemask);
		geode_set_raster_operation(fPtr,XAAGetPatternROP_PM(rop));
	}
}

/**
 * @brief called to fill a rectangle with monochrome pattern of previusly loaded pattern
 * @param[in] pScreenInfo FIXME
 * @param[in] pattern0 same as in setup (to be ignored)
 * @param[in] pattern1 same as in setup (to be ignored)
 * @param[in] x destination x position
 * @param[in] y destination y position
 * @param[in] width destination width
 * @param[in] height destination height
 *
 * Note: As we set the HARDWARE_PATTERN_SCREEN_ORIGIN and HARDWARE_PATTERN_PROGRAMMED_BITS
 * bits, the pattern here are useless and can be ignored
 **/
static void geode_subsequent_8x8_pattern_mono_expand(ScrnInfoPtr pScreenInfo,
		int pattern0, int pattern1, int x, int y, int width, int height)
{
	geode_subsequent_fill_rect_solid(pScreenInfo,x,y,width,height);
}

/* --------------- Line drawing ------------------------------- */

static const unsigned short vector_mode_table[] = {
	VM_MAJOR_INC | VM_MINOR_INC | VM_X_MAJOR,
	VM_MAJOR_INC | VM_MINOR_INC | VM_Y_MAJOR,
	VM_MAJOR_INC | VM_X_MAJOR,
	VM_MINOR_INC | VM_Y_MAJOR,
	VM_MINOR_INC | VM_X_MAJOR,
	VM_MAJOR_INC | VM_Y_MAJOR,
	VM_X_MAJOR,
	VM_Y_MAJOR,
};

/**
 * @brief setup the solid line colour for future line draws
 * @param[in] pScreenInfo screen to draw on
 * @param[in] colour colour value of future lines
 * @param[in] rop X's rop value
 * @param[in] planemask planemask value
 *
 * Note: If planemask is -1 we may need destination data, but never
 * source data. We need source data most of the time, if planemask
 * is not -1!
 **/
static void geode_setup_for_solid_line(ScrnInfoRec *pScreenInfo, int colour, int rop, unsigned int planemask)
{
	GeodeDevRec *fPtr = GEODEDEVPTR( pScreenInfo );
	int components;

	/* xf86Msg(X_INFO,"sfsl: %x, %d\n",colour,rop); */

	if (fPtr->geode_not_native)
		fPtr->GFXsavedColour = geode_expand_8_2_16(colour);
	else
		fPtr->GFXsavedColour = colour;

	if (planemask != 0xFFFFFFFF) {
		fPtr->GFXsavedRop = XAAGetPatternROP_PM(rop);
		components=pattern_components_pm[rop];
	} else {
		fPtr->GFXsavedRop = XAAGetPatternROP(rop);
		components=pattern_components[rop];
	}

	/*
	 * We only setup the blt mode for horizontal/vertical lines, as
	 * they will be drawn by the blt engine instead of the vector engine
	 */
	fPtr->Geode_blt_mode = BM_READ_SRC_NONE | BM_READ_DST_NONE;
	fPtr->Geode_vector_mode = 0;

	if (components & ROP_SRC)	/* source data needed? */
		fPtr->Geode_blt_mode |= BM_READ_SRC_FB;

	if (components & ROP_DST) {	/* destination data needed? */
		fPtr->Geode_vector_mode |= VM_READ_DST_FB;
		fPtr->Geode_blt_mode |= BM_READ_DST_FB0;	/* full buffer */
	}
}

/**
 * @brief render a vector using the specified bresenham parameters
 * @param[in] pScreenInfo screen to draw on
 * @param[in] x1 starting x position
 * @param[in] y1 starting y possition
 * @param[in] absmaj Bresenman absolute major
 * @param[in] absmin Bresenman absolute minor
 * @param[in] err bresenham err term
 * @param[in] len length of the vector interms of pixels
 * @param[in] octant FIXME
 **/
static void geode_subsequent_bresenham_line(ScrnInfoRec *pScreenInfo, int x1,
			int y1, int absmaj, int absmin, int err, int len, int octant)
{
	int axial, init, diag;
	GeodeDevRec *fPtr = GEODEDEVPTR( pScreenInfo );

	/* DEBUGMSG(0, (0, 0, "BLine %d, %d, %d, %d, %d, %d, %d\n",
	x1, y1, absmaj, absmin, err, len, octant)); */

	/* DETERMINE BRESENHAM PARAMETERS */

	axial = absmin * 2;
	init = axial - absmaj;
	diag = init - absmaj;

	/* ADJUST INITIAL ERROR
	 * - Adjust by -1 for certain directions so that the vector
	 * - hits the same pixels when drawn in either direction.
	 * - The Gamma value is assumed to account for the initial
	 * - error adjustment for clipped lines.
	 */

	init += err;

	/* POLL UNTIL ABLE TO WRITE TO THE REGISTERS */
	/* Put off poll for as long as possible (do most calculations first). */

	geode_wait_until_blit_pending(fPtr);
	/* set this data here as we are not sure
	 * they are destroyed by geode_subsequent_hor_vert_line()
	 */
	WRITE_REG16(fPtr->gp_io,GP_PAT_COLOR_0, fPtr->GFXsavedColour);
	WRITE_REG16(fPtr->gp_io,GP_RASTER_MODE, fPtr->GFXsavedRop);

	WRITE_REG16(fPtr->gp_io,GP_DST_XCOOR, x1);	/* signed */
	WRITE_REG16(fPtr->gp_io,GP_DST_YCOOR, y1);	/* signed */

	WRITE_REG16(fPtr->gp_io,GP_VECTOR_LENGTH, len);	/* unsigned */
	WRITE_REG16(fPtr->gp_io,GP_INIT_ERROR, init);	/* unsigned */

	WRITE_REG16(fPtr->gp_io,GP_AXIAL_ERROR, axial);	/* signed */
	WRITE_REG16(fPtr->gp_io,GP_DIAG_ERROR, diag);	/* signed */

	WRITE_REG32(fPtr->gp_io,GP_VECTOR_MODE, (fPtr->Geode_vector_mode | vector_mode_table[octant]));
}

/**
 * @brief FIXME
 * @param[in] pScreenInfo FIXME
 * @param[in] x0 starting x position
 * @param[in] y0 starting y possition
 * @param[in] x1 ending x position
 * @param[in] y1 ending y possition
 * @param[in] flags FIXME
 *
 * Seems to have problems with accuracy!
 **/
#if 0
static void geode_subsequent_solid_two_point_line(ScrnInfoRec *pScreenInfo,
		int x0, int y0, int x1, int y1, int flags)
{
	long dx, dy, dmaj, dmin;
	long axialerr, diagerr, initerr;
	unsigned short vec_flags = 0;
	GeodeDevRec *fPtr = GEODEDEVPTR( pScreenInfo );

	dx = ABS(x1, x0);
	dy = ABS(y1, y0);
	if (dx >= dy) {
		dmaj = dx;
		dmin = dy;
		vec_flags = VM_X_MAJOR;
		if (x1 > x0)
			vec_flags |= VM_MAJOR_INC;
		if (y1 > y0)
			vec_flags |= VM_MINOR_INC;
	} else {
		dmaj = dy;
		dmin = dx;
		vec_flags = VM_Y_MAJOR;
		if (x1 > x0)
			vec_flags |= VM_MINOR_INC;
		if (y1 > y0)
			vec_flags |= VM_MAJOR_INC;
	}

	axialerr = dmin << 1;
	diagerr = (dmin - dmaj) << 1;
	initerr = (axialerr - dmaj);

	if (!(vec_flags & VM_MINOR_INC))
		initerr--;

	geode_wait_until_blit_pending(fPtr);

	WRITE_REG16(fPtr->gp_io,GP_DST_XCOOR, x0);
	WRITE_REG16(fPtr->gp_io,GP_DST_YCOOR, y0);

	WRITE_REG16(fPtr->gp_io,GP_VECTOR_LENGTH, dmaj);
	WRITE_REG16(fPtr->gp_io,GP_INIT_ERROR, initerr);

	WRITE_REG16(fPtr->gp_io,GP_AXIAL_ERROR, axialerr);
	WRITE_REG16(fPtr->gp_io,GP_DIAG_ERROR, diagerr);

	WRITE_REG16(fPtr->gp_io,GP_VECTOR_MODE, (fPtr->Geode_vector_mode | vec_flags));
}
#else
# define geode_subsequent_solid_two_point_line NULL
#endif

/**
 * @brief render a simple vertical or horizontal vector
 * @param[in] pScreenInfo screen to draw on
 * @param[in] x starting x position
 * @param[in] y starting y possition
 * @param[in] len length of line
 * @param[in] dir line direction
 *
 * Its faster to use solid fill instead of the bresenham unit
 **/
static void geode_subsequent_hor_vert_line(ScrnInfoRec *pScreenInfo, int x, int y, int len, int dir)
{
	GeodeDevRec *fPtr = GEODEDEVPTR( pScreenInfo );

	/* xf86Msg(X_INFO,"gshvl\n"); */
	geode_wait_until_blit_pending(fPtr);

	WRITE_REG16(fPtr->gp_io,GP_DST_XCOOR, x);
	WRITE_REG16(fPtr->gp_io,GP_DST_YCOOR, y);
	WRITE_REG16(fPtr->gp_io,GP_HEIGHT, (dir == DEGREES_0) ? 1 : len);
	WRITE_REG16(fPtr->gp_io,GP_WIDTH, (dir == DEGREES_0) ? len : 1);
	WRITE_REG16(fPtr->gp_io,GP_RASTER_MODE, fPtr->GFXsavedRop);
	WRITE_REG16(fPtr->gp_io,GP_PAT_COLOR_0, fPtr->GFXsavedColour);
	WRITE_REG16(fPtr->gp_io,GP_BLIT_MODE, fPtr->Geode_blt_mode);
}

#if 0
void
		GX1SetupForDashedLine(ScrnInfoPtr pScrn, int fg, int bg, int rop,
				      unsigned int planemask, int length,
				      unsigned char *pattern)
{
	int trans = (bg == -1);
	unsigned long *pat = (unsigned long *)pattern;

	/* LOAD PATTERN COLORS AND DATA */

	GFX(set_mono_pattern((unsigned long)bg, (unsigned long)fg,
	    (unsigned long)pat, (unsigned long)pat,
	    (unsigned char)trans));

	/* CHECK IF PLANEMASK IS NOT USED (ALL PLANES ENABLED) */

	if (planemask == (unsigned int)-1) {
		/* USE NORMAL PATTERN ROPs IF ALL PLANES ARE ENABLED */

		GFX(set_raster_operation(XAAGetPatternROP(rop)));
	} else {
		/* SELECT ROP THAT USES SOURCE DATA FOR PLANEMASK */

		GFX(set_solid_source((unsigned long)planemask));
		GFX(set_raster_operation(XAAGetPatternROP_PM(rop)));
	}
}
#endif

/* ------------------------------------------------------------------ */

/**
 * @brief This function is called to syncronise with the graphics
 *        engine and it waits the graphic engine is idle.
 * @param[in] pScreenInfo FIXME
 *
 * This is required before allowing direct access to the framebuffer.
 **/
static void geode_accel_sync(ScrnInfoRec *pScreenInfo)
{
	GeodeDevRec * fPtr = GEODEDEVPTR( pScreenInfo );
	int timeout=1000000;	/* FIXME */

	if (!(READ_REG16(fPtr->gp_io,GP_BLIT_STATUS) & BS_BLIT_BUSY))
		return;

	while (timeout--) {
		/* FIXME use a valuable delay! */
		if (!(READ_REG16(fPtr->gp_io,GP_BLIT_STATUS) & BS_BLIT_BUSY))
			return;
	}
	xf86Msg(X_WARNING, "%s leaving with timeout!\n",__FUNCTION__);
}



/*
 * check valid scratch pad ram
 *
 * TODO: Do realy checks
 */
static int check_for_blit_buffers(GeodeDevRec *fPtr)
{
	unsigned long bb0_base,bb1_base,bb0_pointer,bb1_pointer;

	bb0_base=read_bb0_base();
	bb1_base=read_bb1_base();
	bb0_pointer=read_bb0_pointer();
	bb1_pointer=read_bb1_pointer();

	xf86Msg(X_INFO, "bb0_base=0x%08lX, bb1_base=0x%08lX, bb0_pointer=0x%08lX, bb1_pointer=0x%08lX\n",
		bb0_base,bb1_base,bb0_pointer,bb1_pointer);

	xf86Msg(X_INFO, "First dwords in scratch ram: %08lX - %08lX - %08lX - %08lX\n",
		READ_REG32(fPtr->scratch_io,fPtr->scratch_offset),
		READ_REG32(fPtr->scratch_io,fPtr->scratch_offset+4),
		READ_REG32(fPtr->scratch_io,fPtr->scratch_offset+8),
		READ_REG32(fPtr->scratch_io,fPtr->scratch_offset+12));

	WRITE_REG32(fPtr->scratch_io,fPtr->scratch_offset,0x01020304);
	WRITE_REG32(fPtr->scratch_io,fPtr->scratch_offset+4,0x05060708);
	WRITE_REG32(fPtr->scratch_io,fPtr->scratch_offset+8,0x090A0B0C);
	WRITE_REG32(fPtr->scratch_io,fPtr->scratch_offset+12,0x0D0E0F00);

	xf86Msg(X_INFO, "First dwords in scratch ram: %08lX - %08lX - %08lX - %08lX\n",
		READ_REG32(fPtr->scratch_io,fPtr->scratch_offset),
		READ_REG32(fPtr->scratch_io,fPtr->scratch_offset+4),
		READ_REG32(fPtr->scratch_io,fPtr->scratch_offset+8),
		READ_REG32(fPtr->scratch_io,fPtr->scratch_offset+12));

	return 1;
}

/**
 * @brief Register suppported acceleration features in the X server
 * @param[in] pScreen FIXME
 *
 * This function is called in GX1ScreenInit in geode_driver.c
 * to set the acceleration.
 *
 * @return TRUE on success and FALSE on failure
 **/
int geode_init_acceleration(ScreenRec *pScreen)
{
	ScrnInfoRec *pScreenInfo=xf86Screens[pScreen->myNum];
	GeodeDevRec * fPtr = GEODEDEVPTR( pScreenInfo );
	XAAInfoRec *localRecPtr;

	if (fPtr->scratch_offset == -1) {
		xf86Msg(X_WARNING, "Acceleration not enabled!\n");
		return FALSE;
	}

	/*
	 * 64k colours are Geode's native graphics mode.
	 * 256 indexed colours are not native, so flag it accordingly
	 */
	fPtr->geode_not_native = pScreenInfo->bitsPerPixel == 8;

	switch (fPtr->real_line_length) {
	case 1024:
		fPtr->yshift = 10;
		break;
	case 2048:
		fPtr->yshift = 11;
		break;
	case 4096:
		fPtr->yshift = 12;
		break;
	default:
		xf86Msg(X_WARNING, "Unusable Line Length!\n");
	}

	/* Getting the pointer for acceleration Inforecord */
	fPtr->AccelInfoRec = localRecPtr = XAACreateInfoRec();	/* FIXME error check! */

	/*
	 * The XAA framework needs at least a sync routine!
	 */
	localRecPtr->Sync = geode_accel_sync;

	if (!fPtr->geode_not_native) {
		/* If 16bpp, divide GFXbufferWidthPixels by 2 */
		fPtr->GeodebufferWidthPixels >>= 1;
	}

	localRecPtr->RestoreAccelState = NULL;

	/* SET ACCELERATION FLAGS */
	localRecPtr->Flags = PIXMAP_CACHE /* | OFFSCREEN_PIXMAPS*/ | LINEAR_FRAMEBUFFER;
	localRecPtr->PixmapCacheFlags = DO_NOT_BLIT_STIPPLES;

#if 1
	/*
	 * This hardware supports solid fills
	 */
	localRecPtr->SetupForSolidFill = geode_setup_for_fill_rect;
	localRecPtr->SubsequentSolidFillRect = geode_subsequent_fill_rect_solid;
	localRecPtr->SubsequentSolidFillTrap = NULL;	/* trapezoid */
	localRecPtr->SolidFillFlags = 0;
#endif

#if 1
	/* HOOK 8x8 MonoEXPAND PATTERNS */
	localRecPtr->SetupForMono8x8PatternFill = geode_setup_for_8x8_pattern_mono_expand;
	localRecPtr->SubsequentMono8x8PatternFillRect = geode_subsequent_8x8_pattern_mono_expand;
	/* no trapezium support in this chip */
	localRecPtr->SubsequentMono8x8PatternFillTrap = NULL;
	localRecPtr->Mono8x8PatternFillFlags =	BIT_ORDER_IN_BYTE_MSBFIRST |
						HARDWARE_PATTERN_PROGRAMMED_BITS |
						HARDWARE_PATTERN_SCREEN_ORIGIN;
#endif

#if 1
	/* Color expansion */
	localRecPtr->SetupForColor8x8PatternFill = geode_setup_for_8x8_pattern_colour_expand;
	localRecPtr->SubsequentColor8x8PatternFillRect = geode_subsequent_8x8_pattern_colour_expand;
	/* no trapezium support in this chip */
	localRecPtr->SubsequentColor8x8PatternFillTrap = NULL;
	localRecPtr->Color8x8PatternFillFlags =	BIT_ORDER_IN_BYTE_MSBFIRST |
						SCANLINE_PAD_DWORD |
						HARDWARE_PATTERN_SCREEN_ORIGIN;
#endif

#if 1
	if (check_for_blit_buffers(fPtr)) {
		/* HOOK SCREEN TO SCREEN COPIES
		* * Set flag to only allow copy if transparency is enabled.
		*/
		localRecPtr->SetupForScreenToScreenCopy = geode_setup_for_screen_to_screen_copy;
		localRecPtr->SubsequentScreenToScreenCopy = geode_subsequent_screen_to_screen_copy;
		localRecPtr->ScreenToScreenCopyFlags = 0;
	}
#endif

#if 1
	/* HOOK BRESENHAM SOLID LINES */
	/* Do not hook unless flag can be set preventing use of planemask. */
	localRecPtr->SolidLineFlags = NO_PLANEMASK;
	localRecPtr->SetupForSolidLine = geode_setup_for_solid_line;
	localRecPtr->SubsequentSolidBresenhamLine = geode_subsequent_bresenham_line;
	localRecPtr->SubsequentSolidTwoPointLine = geode_subsequent_solid_two_point_line;
	localRecPtr->SubsequentSolidHorVertLine = geode_subsequent_hor_vert_line;

	localRecPtr->SolidBresenhamLineErrorTermBits = 15;
#endif

#if 1
	localRecPtr->SetupForDashedLine = NULL;
	localRecPtr->DashedLineFlags = 0;
	localRecPtr->DashPatternMaxLength = 0;
	localRecPtr->SubsequentDashedTwoPointLine = NULL;
	localRecPtr->SubsequentDashedBresenhamLine = NULL;

	localRecPtr->DashedBresenhamLineErrorTermBits = 15;
#endif

#if 1
	localRecPtr->SetClippingRectangle = NULL;
	localRecPtr->ClippingFlags = 0;
	localRecPtr->DisableClipping = NULL;
#endif

#if 1
	localRecPtr->SetupForCPUToScreenColorExpandFill = NULL;
	localRecPtr->CPUToScreenColorExpandFillFlags = 0;
	localRecPtr->SubsequentCPUToScreenColorExpandFill = NULL;
	localRecPtr->ColorExpandBase = NULL;
	localRecPtr->ColorExpandRange = 0;
#endif

#if 1
	localRecPtr->SetupForScanlineCPUToScreenColorExpandFill = NULL;
	localRecPtr->ScanlineCPUToScreenColorExpandFillFlags = 0;
	localRecPtr->SubsequentScanlineCPUToScreenColorExpandFill = NULL;
	localRecPtr->SubsequentColorExpandScanline = NULL;
	localRecPtr->NumScanlineColorExpandBuffers = 0;
	localRecPtr->ScanlineColorExpandBuffers = NULL;
#endif

#if 1
	localRecPtr->SetupForScreenToScreenColorExpandFill = NULL;
	localRecPtr->ScreenToScreenColorExpandFillFlags = 0;
	localRecPtr->SubsequentScreenToScreenColorExpandFill = NULL;
#endif

#if 0

   /*
    * ImageWrite.
    *
    * SInce this uses off-screen scanline buffers, it is only of use when
    * complex ROPs are used. But since the current XAA pixmap cache code
    * only works when an ImageWrite is provided, the NO_GXCOPY flag is
    * temporarily disabled.
    */
   if (pGeode->AccelImageWriteBufferOffsets) {

      localRecPtr->ScanlineImageWriteFlags =
	    localRecPtr->ScreenToScreenCopyFlags;
      localRecPtr->ScanlineImageWriteBuffers =
	    pGeode->AccelImageWriteBufferOffsets;
      localRecPtr->NumScanlineImageWriteBuffers = pGeode->NoOfImgBuffers;
      localRecPtr->ImageWriteRange = pGeode->NoOfImgBuffers << gu1_yshift;
      localRecPtr->SetupForScanlineImageWrite =
	    OPTACCEL(GX1SetupForScanlineImageWrite);
      localRecPtr->SubsequentScanlineImageWriteRect =
	    OPTACCEL(GX1SubsequentScanlineImageWriteRect);
      localRecPtr->SubsequentImageWriteScanline =
	    OPTACCEL(GX1SubsequentImageWriteScanline);

      ImgBufOffset = pGeode->AccelImageWriteBufferOffsets[0] - pGeode->FBBase;
      Geodesrcy = ImgBufOffset >> gu1_yshift;

      Geodesrcx = ImgBufOffset & (pGeode->Pitch - 1);
      Geodesrcx /= (pScreenInfo->bitsPerPixel >> 3);
   }
#endif

	return (XAAInit(pScreen, fPtr->AccelInfoRec));
}

/**
 * @brief FIXME
 * @param[in] pScreen FIXME
 *
 * @return TRUE (alsways)
 **/
int geode_exit_acceleration(ScreenRec *pScreen)
{
	ScrnInfoRec *pScreenInfo=xf86Screens[pScreen->myNum];
	GeodeDevRec * fPtr = GEODEDEVPTR( pScreenInfo );

	if (fPtr->AccelInfoRec != NULL)
		XAADestroyInfoRec(fPtr->AccelInfoRec);

	return TRUE;
}

/* end of file geode_accel.c */

#if 0

/* STATIC VARIABLES FOR THIS FILE
 * Used to maintain state between setup and rendering calls.
 */

static int GeodeTransparent;
static int GeodeTransColor;
static int Geodedstx;
static int Geodedsty;
static int Geodesrcx;
static int Geodesrcy;
static int Geodewidth;
static int Geodeheight;
static int GeodeCounter;

#if !defined(STB_X)
static unsigned int GeodeROP = 0;
static unsigned short Geode_blt_mode = 0;
static unsigned short Geode_vector_mode = 0;
static unsigned short Geode_buffer_width = 0;
#endif
static unsigned int gu1_bpp = 0;
static unsigned int gu1_yshift = 1;
static unsigned short GeodebufferWidthPixels;
static unsigned int ImgBufOffset;
static unsigned short Geodebb0Base;
static unsigned short Geodebb1Base;
static XAAInfoRecPtr localRecPtr;


/*----------------------------------------------------------------------------
 * GX1SetupForScanlineImageWrite
 *
 * Description	:This function is used to set up the planemask and raster
 *               for future Bliting functionality.
 *
 * Parameters:
 *		pScreenInfo			:Screen handler pointer having screen information.
 *		rop					:sets the raster operation
 *		transparency_color	:transparency color key.
 *		planemask			:Specifies the value of masking from rop data
 *      bpp					:bits per pixel of the source pixmap
 *		depth				:depth of the source pixmap.
 * Returns		:none
 *
 * Comments		:none
 *  x11perf -putimage10
 *  x11perf -putimage100
 *  x11perf -putimage500
*----------------------------------------------------------------------------
*/

void
GX1SetupForScanlineImageWrite(ScrnInfoPtr pScreenInfo,
			      int rop, unsigned int planemask,
			      int transparency_color, int bpp, int depth)
{
   GFX(set_solid_pattern((unsigned int)planemask));
   /* SET RASTER OPERATION FOR USING PATTERN AS PLANE MASK */
   GFX(set_raster_operation(XAAGetCopyROP_PM(rop)));
   /* SAVE TRANSPARENCY FLAG */
   GeodeTransparent = (transparency_color == -1) ? 0 : 1;
   GeodeTransColor = transparency_color;
}

/*----------------------------------------------------------------------------
 * GX1SubsequentScanlineImageWriteRect
 *
 * Description	:This function is used to set up the x,y corordinates and width
 *               &height for future Bliting functionality.
 *
 * Parameters:
 *	pScreenInfo :Screen handler pointer having screen information.
 *      x		:destination x
 *	    y		:destination y
 *		w		:Specifies the width of the rectangle to be  copied
 *		h		:Specifies the height of the rectangle to be  copied
 *
 * Returns		:none
 *
 * Comments		:none
*----------------------------------------------------------------------------
*/
void
GX1SubsequentScanlineImageWriteRect(ScrnInfoPtr pScreenInfo,
				    int x, int y, int w, int h, int skipleft)
{

   Geodedstx = x;
   Geodedsty = y;
   Geodewidth = w;
   Geodeheight = h;
   GeodeCounter = 0;
}

/*----------------------------------------------------------------------------
 * GX1SubsquentImageWriteScanline
 *
 * Description	:This function is called to
 *               BLT  using the previously  specified planemask,raster
 *               operation and   transparency flag
 *
 * Parameters.
 * 	pScreenInfo :Screen handler pointer having screen information.
 *
 * Returns		:none
 *
 * Comments		:The patterns specified is ignored inside the function
 * Sample application uses (non-transparent):
 *   - Moving windows.
 *   - x11perf: scroll tests (-scroll500).
 *   - x11perf: copy from window to window (-copywinwin500).
 *
 * No application found using transparency.
*----------------------------------------------------------------------------
*/

void
GX1SubsequentImageWriteScanline(ScrnInfoPtr pScreenInfo, int bufno)
{
   GeodePtr pGeode = GEODEPTR(pScreenInfo);
   int blt_height = 0;
   char blit = FALSE;

   GeodeCounter++;

   if ((Geodeheight <= pGeode->NoOfImgBuffers) &&
       (GeodeCounter == Geodeheight)) {
      blit = TRUE;
      blt_height = Geodeheight;
} else if ((Geodeheight > pGeode->NoOfImgBuffers)
	      && (GeodeCounter == pGeode->NoOfImgBuffers)) {
      blit = TRUE;
      Geodeheight -= pGeode->NoOfImgBuffers;
      blt_height = pGeode->NoOfImgBuffers;
} else
      return;

   if (blit) {
      blit = FALSE;

      GeodeCounter = 0;

      if (GeodeTransparent) {
	 /* CALL ROUTINE FOR TRANSPARENT SCREEN TO SCREEN BLT
	  * * Should only be called for the "copy" raster operation.
	 */
	 GFX(screen_to_screen_xblt((unsigned short)Geodesrcx,
				   (unsigned short)Geodesrcy,
				   (unsigned short)Geodedstx,
				   (unsigned short)Geodedsty,
				   (unsigned short)Geodewidth,
				   (unsigned short)blt_height,
				   GeodeTransColor));
} else {
	 /* CALL ROUTINE FOR NORMAL SCREEN TO SCREEN BLT */
	 GFX(screen_to_screen_blt((unsigned short)Geodesrcx,
				  (unsigned short)Geodesrcy,
				  (unsigned short)Geodedstx,
				  (unsigned short)Geodedsty,
				  (unsigned short)Geodewidth,
				  (unsigned short)blt_height));
}
      Geodedsty += blt_height;
      GFX(wait_until_idle());
}
}




#if !defined(STB_X)

/*----------------------------------------------------------------------------
 * OPTGX1SetupForScanlineImageWrite
 *
 * Description	:This function is used to set up the planemask and raster
 *               for future Bliting functionality.(non durango version)
 *
 * Parameters:
 *		pScreenInfo			:Screen handler pointer having screen information.
 *      rop					:sets the raster operation
 *		transparency_color	:tobeadded
 *		planemask			:Specifies the value of masking from rop data
 *      bpp					:bits per pixel of the source pixmap
 *		depth				:depth of the source pixmap.
 * Returns		:none
 *
 * Comments		:none
*----------------------------------------------------------------------------
*/
void
OPTGX1SetupForScanlineImageWrite(ScrnInfoPtr pScreenInfo,
				 int rop, unsigned int planemask,
				 int transparency_color, int bpp, int depth)
{

   OPTGX1SetupForScreenToScreenCopy(pScreenInfo,
				    0, 0, rop, planemask, transparency_color);
}

/*----------------------------------------------------------------------------
 * OPTGX1SubsequentScanlineImageWriteRect
 *
 * Description	:This function is used to set up the x,y corordinates and width
 *                &height for future Bliting functionality.(non durango version)
 *
 * Parameters:
 *	pScreenInfo :Screen handler pointer having screen information.
 *      x		:destination x
 *	    y		:destination y
 *		w		:Specifies the width of the rectangle to be  copied
 *		h		:Specifies the height of the rectangle to be  copied
 * Returns		:none
 *
 * Comments		:none
 *----------------------------------------------------------------------------
*/
void
OPTGX1SubsequentScanlineImageWriteRect(ScrnInfoPtr pScreenInfo,
				       int x, int y, int w, int h,
				       int skipleft)
{

   Geodedstx = x;
   Geodedsty = y;
   Geodewidth = w;
   Geodeheight = h;
   GeodeCounter = 0;
}

/*----------------------------------------------------------------------------
 * OPTGX1SubsquentImageWriteScanline
 *
 * Description	:This function is called to
 *               BLT  using the previously  specified planemask,raster
 *               operation and transparency flag(non durango version)
 *
 * Parameters.
 * 	pScreenInfo :Screen handler pointer having screen information.
 *
 * Returns		:none
 *
 * Comments		:The patterns specified is ignored inside the function
 * Sample application uses (non-transparent):
 *   - Moving windows.
 *   - x11perf: scroll tests (-scroll500).
 *   - x11perf: copy from window to window (-copywinwin500).
 *
 * No application found using transparency.
*----------------------------------------------------------------------------
*/

void
OPTGX1SubsequentImageWriteScanline(ScrnInfoPtr pScreenInfo, int bufno)
{
   GeodePtr pGeode = GEODEPTR(pScreenInfo);
   int blt_height = 0;
   char blit = FALSE;

   GeodeCounter++;

   if ((Geodeheight <= pGeode->NoOfImgBuffers) &&
       (GeodeCounter == Geodeheight)) {
      blit = TRUE;
      blt_height = Geodeheight;
} else if ((Geodeheight > pGeode->NoOfImgBuffers)
	      && (GeodeCounter == pGeode->NoOfImgBuffers)) {
      blit = TRUE;
      Geodeheight -= pGeode->NoOfImgBuffers;
      blt_height = pGeode->NoOfImgBuffers;
} else
      return;

   if (blit) {
      blit = FALSE;
      GeodeCounter = 0;
      OPTGX1SubsequentScreenToScreenCopy(pScreenInfo,
					 Geodesrcx, Geodesrcy, Geodedstx,
					 Geodedsty, Geodewidth, blt_height);
      Geodedsty += blt_height;
      GFX_WAIT_BUSY;
}
}

#endif
#endif
