/**
 * @file geode-fb.c
 * @brief Native driver for the Geode GX1 processor and chipset
 * @author Juergen Beisert <jbeisert@netscape.net>
 *
 * This driver bases on Xorg's fbdev
 */

#undef DEBUG

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <string.h>

/* all driver need this */
#include <xf86.h>
#include <xf86_OSproc.h>

/* All drivers initialising the SW cursor need this */
#include <mipointer.h>

/* All drivers implementing backing store need this */
/* #include "mibstore.h" */

#include <micmap.h>
#include <colormapst.h>
#include <xf86cmap.h>

/* for visuals */
#include <fb.h>

#include <xf86Resources.h>
#include <xf86RAC.h>

#include <xf86cmap.h>

#include <globals.h>
#define DPMS_SERVER
#include <X11/extensions/dpms.h>

#include <fcntl.h>
#include <errno.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <math.h>	/* for gamma correction calculation */

#include "geode-fb.h"

/*
 What is the difference between bitsPerPixel and depth?

 The depth is the number of bits in each pixel that are significant
 in determining the pixel's value. The bitsPerPixel is the total
 size occupied by each pixel, including bits that are not used.
*/

/**
 * @brief currently not used
 *
 * to be removed
 **/
enum {
	FBDEV_ROTATE_NONE = 0,
	FBDEV_ROTATE_CW = 270,
	FBDEV_ROTATE_UD = 180,
	FBDEV_ROTATE_CCW = 90
};

/**
 * @brief Supported options as constants
 */
typedef enum {
	OPTION_SHADOW_FB,
	OPTION_ROTATE,	/* to be removed */
	OPTION_ACCEL,
	OPTION_FBDEV,
	OPTION_COMPRESS
} FBDevOpts;

/**
 * @brief List of options we can handle here in this driver
 *
 * Used by FBDevAvailableOptions
 */
static const OptionInfoRec FBDevOptions[] =
{
	{
		.token	=	OPTION_SHADOW_FB,
		.name	=	"ShadowFB",
		.type	=	OPTV_BOOLEAN,
		.value	=	{0},
		.found	=	FALSE
	},
	{ /* to be removed */
		.token	=	OPTION_ROTATE,
		.name	=	"Rotate",
		.type	=	OPTV_STRING,
		.value	=	{0},
		.found	=	FALSE
	},
	{
		.token	=	OPTION_ACCEL,
		.name	=	"Accel",
		.type	=	OPTV_BOOLEAN,
		.value	=	{0},
		.found	=	FALSE
	},
	{
		.token	=	OPTION_COMPRESS,
		.name	=	"Compression",
		.type	=	OPTV_BOOLEAN,
		.value	=	{0},
		.found	=	FALSE
	},
	{
		.token	=	OPTION_FBDEV,
		.name	=	"fbdev",
		.type	=	OPTV_STRING,
		.value	=	{0},
		.found	=	FALSE
	},
	{ .token=-1, .name=NULL, .type=OPTV_NONE, .value={0}, .found=FALSE }	/* last entry marker */
};


/**
 * @brief required symbols from the framebuffer library
 **/
const char *fbSymbols[] =
{
	"fbScreenInit",
	"fbPictureInit",
	NULL
};

#if 0
/**
 * @brief required symbols from the shadow framebuffer library
 **/
const char *shadowSymbols[] =
{
	"shadowAdd",
	"shadowInit",
	"shadowSetup",
	"shadowUpdatePacked",
	"shadowUpdatePackedWeak",
	"shadowUpdateRotatePacked",
	"shadowUpdateRotatePackedWeak",
	NULL
};
#endif

/**
 * @brief required symbols from the xaa library (for acceleration)
 **/
static const char *xaaSymbols[] = {
	"XAADestroyInfoRec",
	"XAACreateInfoRec",
	"XAAInit",
	NULL
};

/**
 * @brief required symbols from the ramdac library (for hardware cursor)
 **/
static const char *ramdacSymbols[] = {
	"xf86CreateCursorInfoRec",
	"xf86DestroyCursorInfoRec",
	"xf86InitCursor",
	NULL
};

/* -------------------------------------------------------------------- */

/**
 * @brief This function returns the options set
 * @param chipid This will identify the chipset.
 * @param busid This will identify the PCI busid
 *
 * @return pointer to FBDevOptions
 */
const OptionInfoRec *FBDevAvailableOptions( int chipid, int busid )
{
	return FBDevOptions;
}

/**
 * @brief Register the symbols required to run this driver
 **/
int geode_gx1_register_symbols(void)
{
	LoaderRefSymLists( fbSymbols, ramdacSymbols, xaaSymbols, NULL );
	return TRUE;
}

/**
 * @brief free remaining resources
 * @param[in] scrnIndex FIXME
 * @param[in] flags FIXME
 **/
static void geode_free_screen(int scrnIndex, int flags)
{
	ScrnInfoRec *pScreenInfo=xf86Screens[scrnIndex];

	TRACE_ENTER(__FUNCTION__);
	xfree(pScreenInfo->driverPrivate);
	pScreenInfo->driverPrivate = NULL;
}

#ifdef DEBUG
static void print_fbdev_mode(char *txt, struct fb_var_screeninfo *var)
{
	ErrorF( "fbdev %s mode:\t%d   %d %d %d %d   %d %d %d %d   %d %d:%d:%d\n",
		txt,var->pixclock,
		var->xres, var->right_margin, var->hsync_len, var->left_margin,
		var->yres, var->lower_margin, var->vsync_len, var->upper_margin,
		var->bits_per_pixel,
		var->red.length, var->green.length, var->blue.length);
}

static void print_xfree_mode(char *txt, DisplayModePtr mode)
{
	ErrorF( "xfree %s mode:\t%d   %d %d %d %d   %d %d %d %d\n",
		txt,mode->Clock,
		mode->HDisplay, mode->HSyncStart, mode->HSyncEnd, mode->HTotal,
		mode->VDisplay, mode->VSyncStart, mode->VSyncEnd, mode->VTotal);
}
#endif

/**
 * @brief Open a given frambuffer device, or try one given in FRAMEBUFFER
 * @param[in] dev Device node path and name (may be NULL)
 *
 * dev is from our driver's screen section
 *   Option "fbdev" "/dev/fb0"
 * and will be used when given.
 * If no name was given the environment variable FRAMEBUFFER
 * will be tried and "/dev/fb0" as a fall back if no environment
 * was set.
 *
 * Note: We cannot output any messages here
 *
 * @return Device handle if successfull or -1 in case of error
 **/
static int geode_fb_open(const char *dev)
{
	char *default_dev;
	int    fd;

	TRACE_ENTER(__FUNCTION__);
	/* try argument (from XF86Config) first */
	if (dev) {
		fd = open(dev,O_RDWR,0);
	} else {
		xf86Msg(X_WARNING, "Not framebuffer device node name given. Falling back to defaults.\n");
		/* second: environment variable */
		default_dev = getenv("FRAMEBUFFER");
		if ((NULL == default_dev) || ((fd = open(default_dev,O_RDWR,0)) == -1)) {
			/* last try: default device */
			fd = open("/dev/fb0",O_RDWR,0);
		}
	}
	if (fd == -1)
		return fd;

	return fd;
}

/* -------------------------------------------------------------------- */

/**
 * @brief Check if a geode framebuffer device node is present
 * @param[in] device Framebuffer device node name
 *
 * Also test if we can query the framebuffer driver for all
 * information we rely on.
 *
 * @return TRUE if device found, FALSE elsewhere
 **/
static Bool geode_HW_probe(const char *device)
{
	int fd;
	Bool rc=TRUE;
	struct geodefb_ioctl geode_fb;
	struct fb_fix_screeninfo fix;

	TRACE_ENTER(__FUNCTION__);

	fd = geode_fb_open(device);

	if (-1 == fd)
		return FALSE;	/* does not exist */

	/*
	 * Do some basic tests
	 */
	if (ioctl(fd,FBIOGET_HARDWARE_PARAM,(void*)(&geode_fb)) == -1) {
		rc=FALSE;	/* this is not our frame buffer driver */
		xf86Msg(X_NOT_IMPLEMENTED, "No framebffer driver to colaborate\n");
		goto on_error_leave;
	}
	if (ioctl(fd,FBIOGET_FSCREENINFO,(void*)(&fix)) == -1) {
		rc = FALSE;
		xf86Msg(X_NOT_IMPLEMENTED, "No framebffer driver to colaborate\n");
		goto on_error_leave;
	}
	/*
	 * we need all addresses with usable values!
	 */
	if (geode_fb.dc_address == NULL ||
	    geode_fb.gp_address == NULL ||
	    geode_fb.sram_address == NULL ||
	    geode_fb.screen_start == NULL ||
	    fix.mmio_start == 0) {
		rc=FALSE;
		xf86Msg(X_NOT_IMPLEMENTED, "Framebffer driver to old, can't colaborate\n");
		goto on_error_leave;
	}

on_error_leave:

	close(fd);
	return rc;
}

/**
 * @brief Collect information about the hardware
 * @param[in] pScrn FIXME
 * @param[in] pPci FIXME
 * @param[in] device Name of the framebuffer device node to open
 **/
static Bool geode_hw_init(ScrnInfoRec *pScrn, pciVideoPtr pPci, char *device)
{
	struct geodefb_ioctl geode_fb;
	GeodeDevRec *fPtr=GEODEDEVPTR(pScrn);

	TRACE_ENTER(__FUNCTION__);

	/* open device */
	fPtr->fd = geode_fb_open(device);
	if (fPtr->fd == -1) {
		xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
			"Failed to open framebuffer device, consult warnings"
			" and/or errors above for possible reasons\n"
			"\t(you may have to look at the server log to see"
			" warnings)\n");
		return FALSE;
	}
	/*
	 * get some information about the hardware we are running on
	 * most of this information we got from the framebuffer driver
	 */
	if (ioctl(fPtr->fd,FBIOGET_HARDWARE_PARAM,(void*)(&geode_fb)) == -1) {
		xf86DrvMsg(pScrn->scrnIndex, X_ERROR,"ioctl FBIOGET_HARDWARE_PARAM: %s\n",strerror(errno));
		xf86DrvMsg(pScrn->scrnIndex, X_INFO,"Update your kernel!\n");
	}
	else {
		fPtr->dc_registers=(unsigned long)geode_fb.dc_address;
		fPtr->gp_registers=(unsigned long)geode_fb.gp_address;
		fPtr->scratch_pad_ram=(unsigned long)geode_fb.sram_address;
		fPtr->scratch_offset=geode_fb.sram_offset;
		fPtr->video_mem=(unsigned long)geode_fb.screen_start;
	}

	/* get current fb device settings */
	if (ioctl(fPtr->fd,FBIOGET_FSCREENINFO,(void*)(&fPtr->fix)) == -1) {
		xf86DrvMsg(pScrn->scrnIndex, X_ERROR,"ioctl FBIOGET_FSCREENINFO: %s\n",strerror(errno));
		return FALSE;
	}
	if (ioctl(fPtr->fd,FBIOGET_VSCREENINFO,(void*)(&fPtr->var)) == -1) {
		xf86DrvMsg(pScrn->scrnIndex, X_ERROR,"ioctl FBIOGET_VSCREENINFO: %s\n",strerror(errno));
		return FALSE;
	}

	if (fPtr->fix.mmio_start == 0) {
		xf86DrvMsg(pScrn->scrnIndex, X_ERROR,"Kernel to old!\n");
		return FALSE;
	}
	/*
	 * The standard IOCTLs do not support Geode's
	 * specific behaviour. So we fakte the line length
	 * here, until a new video mode is set and the framebuffer
	 * driver reports the correct length
	*/
	fPtr->real_line_length = fPtr->fix.line_length;

	/* this is system specific and not *very* relevant in this driver (but used!) */
	fPtr->pagesize = sysconf(_SC_PAGESIZE) - 1;

	return TRUE;
}

/**
 * @brief FIXME
 **/
static void xfree2fbdev_fblayout(ScrnInfoRec *pScrn, struct fb_var_screeninfo *var)
{
	var->xres_virtual   = pScrn->virtualX;
	var->yres_virtual   = pScrn->virtualY;
	var->bits_per_pixel = pScrn->bitsPerPixel;
	var->red.length     = pScrn->weight.red;
	var->green.length   = pScrn->weight.green;
	var->blue.length    = pScrn->weight.blue;
}

/**
 * @brief FIXME
 **/
static void xfree2fbdev_timing(DisplayModeRec *mode, struct fb_var_screeninfo *var)
{
	var->xres = mode->HDisplay;
	var->yres = mode->VDisplay;
	if (var->xres_virtual < var->xres)
		var->xres_virtual = var->xres;
	if (var->yres_virtual < var->yres)
		var->yres_virtual = var->yres;
	var->xoffset = var->yoffset = 0;
	var->pixclock = mode->Clock ? 1000000000/mode->Clock : 0;
	var->right_margin = mode->HSyncStart-mode->HDisplay;
	var->hsync_len = mode->HSyncEnd-mode->HSyncStart;
	var->left_margin = mode->HTotal-mode->HSyncEnd;
	var->lower_margin = mode->VSyncStart-mode->VDisplay;
	var->vsync_len = mode->VSyncEnd-mode->VSyncStart;
	var->upper_margin = mode->VTotal-mode->VSyncEnd;
	var->sync = 0;
	if (mode->Flags & V_PHSYNC)
		var->sync |= FB_SYNC_HOR_HIGH_ACT;
	if (mode->Flags & V_PVSYNC)
		var->sync |= FB_SYNC_VERT_HIGH_ACT;
	if (mode->Flags & V_PCSYNC)
		var->sync |= FB_SYNC_COMP_HIGH_ACT;
	if (mode->Flags & V_INTERLACE)
		var->vmode = FB_VMODE_INTERLACED;
	else if (mode->Flags & V_DBLSCAN)
		var->vmode = FB_VMODE_DOUBLE;
	else
		var->vmode = FB_VMODE_NONINTERLACED;
}

/**
 * @brief Save the current hardware state
 * @param[in] pScrn FIXME
 *
 * Save the hardware state if we will fo some strange
 * things with it the framebuffer driver can't handle
 **/
static void geode_save_HW_state(ScrnInfoRec *pScrn)
{
	TRACE_ENTER(__FUNCTION__);
#if 0
	GeodeDevRec * fPtr = GEODEDEVPTR(pScrn);

	TRACE_ENTER(__FUNCTION__);
	if (0 != ioctl(fPtr->fd,FBIOGET_VSCREENINFO,(void*)(&fPtr->saved_var)))
		xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
			   "FBIOGET_VSCREENINFO: %s\n", strerror(errno));
#endif
}

/**
 * @brief Restore hardware state
 * @param[in] pScrn FIXME
 *
 * Restore the hardware state if we did some strange
 * things with it the framebuffer driver can't handle
 **/
static void geode_restore_HW_state(ScrnInfoRec *pScrn)
{
	TRACE_ENTER(__FUNCTION__);
#if 0
	GeodeDevRec * fPtr = GEODEDEVPTR(pScrn);

	if (0 != ioctl(fPtr->fd,FBIOPUT_VSCREENINFO,(void*)(&fPtr->saved_var)))
		xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
			   "FBIOPUT_VSCREENINFO: %s\n", strerror(errno));
#endif
}

/**
 * @brief Display Power Management Support
 * @param[in] pScrn FIXME
 * @param[in] mode New mode to enter
 * @param[in] flags FIXME
 **/
static void geode_DPMS_feature(ScrnInfoRec *pScrn, int mode, int flags)
{
	GeodeDevRec * fPtr = GEODEDEVPTR(pScrn);
	unsigned long fbmode;

	TRACE_ENTER(__FUNCTION__);
	if (!pScrn->vtSema)
		return;

	switch (mode) {
	case DPMSModeOn:
		fbmode = FB_BLANK_UNBLANK;	/* active */
		break;
	case DPMSModeStandby:
		fbmode = FB_BLANK_VSYNC_SUSPEND;
		break;
	case DPMSModeSuspend:
		fbmode = FB_BLANK_HSYNC_SUSPEND;
		break;
	case DPMSModeOff:
		fbmode = FB_BLANK_POWERDOWN;
		break;
	default:
		xf86DrvMsg(pScrn->scrnIndex, X_ERROR,"FBIOBLANK: Unknown mode %d\n", mode);
		return;
	}

	if (-1 == ioctl(fPtr->fd, FBIOBLANK, (void *)fbmode))
		xf86DrvMsg(pScrn->scrnIndex, X_ERROR,"FBIOBLANK: %s\n", strerror(errno));
}

/**
 * @brief setup new gamma value into the hardware
 * @param[in] snum screen number
 * @param[in] gamma new gamma values
 *
 * Note: This function will only be used from the xv extension.
 **/
#if 0
int geode_change_gamma(int snum, Gamma gamma)
{
	ScrnInfoPtr pScrn = xf86Screens[snum];
	GeodeDevRec * fPtr = GEODEDEVPTR(pScrn);
	LOCO colours[256];
	int cindex;
	double x,y[3];

	TRACE_ENTER(__FUNCTION__);

	if (gamma.red < GAMMA_MIN || gamma.red > GAMMA_MAX ||
		gamma.green < GAMMA_MIN || gamma.green > GAMMA_MAX ||
		gamma.blue < GAMMA_MIN || gamma.blue > GAMMA_MAX)
	return BadValue;

	pScrn->gamma.red = gamma.red;
	pScrn->gamma.green = gamma.green;
	pScrn->gamma.blue = gamma.blue;
	/*
	 * what is the gamma function?
	 */
	y[0]=1.0/pScrn->gamma.red;
	y[1]=1.0/pScrn->gamma.green;
	y[2]=1.0/pScrn->gamma.blue;
	for(cindex=0;cindex<256;cindex++) {
		x=(double)cindex/255.0;
		colours[cindex].red=(unsigned short)(pow(x,y[0]) * 255.5);
		colours[cindex].green=(unsigned short)(pow(x,y[1]) * 255.5);
		colours[cindex].blue=(unsigned short)(pow(x,y[2]) * 255.5);
	}
	/*
	 * load the table into the gamma correction table
	 */
	return geode_setup_gamma_table(fPtr,colours);
}
#else
# define geode_change_gamma NULL
#endif
/**
 * @brief FIXME
 **/
static void fbdevHWAdjustFrame(int scrnIndex, int x, int y, int flags)
{
	ScrnInfoPtr pScrn = xf86Screens[scrnIndex];
	GeodeDevRec * fPtr = GEODEDEVPTR(pScrn);

	TRACE_ENTER(__FUNCTION__);
	if ( x < 0 || x + fPtr->var.xres > fPtr->var.xres_virtual ||
		    y < 0 || y + fPtr->var.yres > fPtr->var.yres_virtual )
		return;

	fPtr->var.xoffset = x;
	fPtr->var.yoffset = y;
	if (-1 == ioctl(fPtr->fd,FBIOPAN_DISPLAY,(void*)&fPtr->var))
		xf86DrvMsgVerb(scrnIndex, X_WARNING, 5,"FBIOPAN_DISPLAY: %s\n", strerror(errno));
}

/**
 * @brief FIXME
 **/
static void fbdevHWLoadPalette(ScrnInfoPtr pScrn, int numColors, int *indices,
			       LOCO *colors, VisualPtr pVisual)
{
	GeodeDevRec * fPtr = GEODEDEVPTR(pScrn);
	struct fb_cmap cmap;
	unsigned short red,green,blue;
	int i;

	TRACE_ENTER(__FUNCTION__);
	cmap.len   = 1;
	cmap.red   = &red;
	cmap.green = &green;
	cmap.blue  = &blue;
	cmap.transp = NULL;
	for (i = 0; i < numColors; i++) {
		cmap.start = indices[i];
		red   = (colors[indices[i]].red   << 8) | colors[indices[i]].red;
		green = (colors[indices[i]].green << 8) | colors[indices[i]].green;
		blue  = (colors[indices[i]].blue  << 8) | colors[indices[i]].blue;
		if (-1 == ioctl(fPtr->fd,FBIOPUTCMAP,(void*)&cmap))
			xf86DrvMsg(pScrn->scrnIndex, X_ERROR,"FBIOPUTCMAP: %s\n", strerror(errno));
	}
}

/**
 * @brief Blank/unblank the screen
 * @param[in] pScreen FIXME
 * @param[in] mode New mode to enter
 *
 * @return TRUE if mode is reached, FALSE if not possible
 **/
static Bool geode_save_screen(ScreenPtr pScreen, int mode)
{
	ScrnInfoPtr pScrn = xf86Screens[pScreen->myNum];
	GeodeDevRec * fPtr = GEODEDEVPTR(pScrn);
	unsigned long unblank;

	TRACE_ENTER(__FUNCTION__);
	if (!pScrn->vtSema)
		return TRUE;

	unblank = xf86IsUnblank(mode);

	if (-1 == ioctl(fPtr->fd, FBIOBLANK, (void *)(1-unblank))) {
		xf86DrvMsg(pScrn->scrnIndex, X_ERROR,"FBIOBLANK: %s\n", strerror(errno));
		return FALSE;
	}

	return TRUE;
}

/**
 * @brief Unmap all IO areas
 * @param[in] pScrn FIXME
 *
 * @return TRUE on success, FALSE, elsewhere
 **/
static Bool geode_unmap_IO_memory(ScrnInfoPtr pScrn)
{
	GeodeDevRec * fPtr = GEODEDEVPTR(pScrn);

	TRACE_ENTER(__FUNCTION__);

	if (fPtr->vmem_io != NULL) {
		xf86UnMapVidMem(pScrn->scrnIndex,fPtr->vmem_io,fPtr->fbmem_len);
		fPtr->vmem_io = NULL;
	}

	if (fPtr->pci_io != NULL) {
		xf86UnMapVidMem(pScrn->scrnIndex,fPtr->pci_io,fPtr->fix.mmio_len);
		fPtr->pci_io = NULL;
	}

	if (fPtr->dc_io != NULL) {
		xf86UnMapVidMem(pScrn->scrnIndex,fPtr->dc_io,GEODE_DC_WINDOW_SIZE);
		fPtr->dc_io = NULL;
	}

	if (fPtr->gp_io != NULL) {
		xf86UnMapVidMem(pScrn->scrnIndex,fPtr->gp_io,GEODE_GP_WINDOW_SIZE);
		fPtr->gp_io = NULL;
	}

	if (fPtr->scratch_io != NULL) {
		xf86UnMapVidMem(pScrn->scrnIndex,fPtr->scratch_io,GEODE_SCRATCH_WINDOW_SIZE);
		fPtr->scratch_io = NULL;
	}
	return TRUE;
}

/**
 * @brief map all required memories to access from user space
 * @param[in] pScrn screen
 * @param[in] fPtr Pointer to driver's private data
 *
 * @return TRUE on success
 **/
static Bool geode_map_IO_memory(ScrnInfoRec *pScrn, GeodeDevRec *fPtr)
{
	TRACE_ENTER(__FUNCTION__);

	/* FIXME What the hell...*/
	fPtr->fboff = fPtr->video_mem & fPtr->pagesize;
	fPtr->fbmem_len = (fPtr->fboff+fPtr->fix.smem_len + fPtr->pagesize) & ~fPtr->pagesize;

	/*
	 * map video memory and various IO spaces
	 */
	xf86DrvMsg(pScrn->scrnIndex, X_INFO,"Mapping video memory at 0x%08lX (size=%dkiB)\n",
		   fPtr->video_mem,fPtr->fbmem_len>>10);
	fPtr->vmem_io = xf86MapVidMem(pScrn->scrnIndex,VIDMEM_FRAMEBUFFER,fPtr->video_mem,fPtr->fbmem_len);

	if (fPtr->vmem_io == NULL) {
		xf86DrvMsg(pScrn->scrnIndex, X_ERROR,"Could not map video memory area\n");
		return FALSE;
	}

	xf86DrvMsg(pScrn->scrnIndex, X_INFO,"Mapping PCI registers at 0x%08lX (size=%d)\n",
		   fPtr->fix.mmio_start,fPtr->fix.mmio_len);
	fPtr->pci_io = xf86MapVidMem(pScrn->scrnIndex,VIDMEM_MMIO,fPtr->fix.mmio_start,fPtr->fix.mmio_len);

	xf86DrvMsg(pScrn->scrnIndex, X_INFO,"Mapping DC registers at 0x%08lX (size=%d)\n",
		   fPtr->dc_registers,GEODE_DC_WINDOW_SIZE);
	fPtr->dc_io = xf86MapVidMem(pScrn->scrnIndex,VIDMEM_MMIO,fPtr->dc_registers,GEODE_DC_WINDOW_SIZE);

	xf86DrvMsg(pScrn->scrnIndex, X_INFO,"Mapping GP registers at 0x%08lX (size=%d)\n",
		   fPtr->gp_registers,GEODE_GP_WINDOW_SIZE);
	fPtr->gp_io = xf86MapVidMem(pScrn->scrnIndex,VIDMEM_MMIO,fPtr->gp_registers,GEODE_GP_WINDOW_SIZE);

	if (fPtr->accel) {
		xf86DrvMsg(pScrn->scrnIndex, X_INFO,"Mapping scratch pad RAM at 0x%08lX (size=%d)\n",
			   fPtr->scratch_pad_ram,GEODE_SCRATCH_WINDOW_SIZE);
		fPtr->scratch_io = xf86MapVidMem(pScrn->scrnIndex,VIDMEM_MMIO,fPtr->scratch_pad_ram,GEODE_SCRATCH_WINDOW_SIZE);
	}
	else
		fPtr->scratch_io=NULL;

	if ((fPtr->pci_io == NULL) || (fPtr->dc_io == NULL) || (fPtr->gp_io == NULL)) {
		xf86DrvMsg(pScrn->scrnIndex, X_ERROR,"Could not map IO areas\n");
		return FALSE;
	}

	if (fPtr->accel && (fPtr->scratch_io == NULL)) {
		xf86DrvMsg(pScrn->scrnIndex, X_ERROR,"Could not map Scratch IO area\n");
		return FALSE;
	}

	pScrn->memPhysBase = fPtr->video_mem & ~fPtr->pagesize;
	pScrn->fbOffset = fPtr->fboff;

	/* FIXME: Should we unmap here in case of failure? */
	return TRUE;
}

/**
 * @brief calculate offscreen memory for cursor and bitblit
 * @param[in] scrnIndex to print something
 * @param[in] fPtr pointer to our private data
 * @param[in] width horizontal pixel
 * @param[in] height vertical lines
 * @param[in] bpp bits per pixel (physical)
 *
 * This is part of the acceleration
 *
 * @return FALSE if there is no space to place cursor images
 **/
static int geode_calc_offscreen_memories(int scrnIndex,GeodeDevRec *fPtr, int width, int height, int bpp)
{
	unsigned int size;
	/*
	 * at first lets see if some remaining space is
	 * between end of visible screen and end of video memory
	*/
	size = fPtr->real_line_length * height;
	/* check for cursor only */
	if ((size + 0x200) < fPtr->fbmem_len) {
	/* locate both buffers at the end of the available memory */
		fPtr->cursor_offset_1 = fPtr->fbmem_len-0x100;
		fPtr->cursor_offset_2 = fPtr->fbmem_len-0x200;
		xf86DrvMsg(scrnIndex, X_INFO,"Locating cursor images at the end (%lx,%lx)\n",fPtr->cursor_offset_1,fPtr->cursor_offset_2);
	} else {
	/* there is no space left at the end of the visual buffer.
	 * This is valid in the 1280x1024-16 mode when acceleration is
	 * enabled (real_line_length will be 4096 bytes in this case)
	 * Next we try to place the cursor between two visual lines.
	 * This only works, if compression is disabled, otherwise this
	 * space is used by compression itself (compression needs up
	 * to 256 bytes per line)!
	 */
		if ((fPtr->real_line_length - (bpp*width/8)) >= 0x100) {
			/*
			 * all right there is space between two lines
			 * Locate the images in the last two lines
			 */
			fPtr->cursor_offset_1 = fPtr->fbmem_len-0x100;
			fPtr->cursor_offset_2 = fPtr->real_line_length * (height-1) - 0x100;
			xf86DrvMsg(scrnIndex, X_INFO,"Locating cursor between lines (%lx,%lx)\n",fPtr->cursor_offset_1,fPtr->cursor_offset_2);
		} else {
			xf86DrvMsg(scrnIndex, X_ERROR,"Cannot place cursor images\n");
			return FALSE;
		}
	}
	fPtr->cursor_1 = fPtr->vmem_io+fPtr->cursor_offset_1;
	fPtr->cursor_2 = fPtr->vmem_io+fPtr->cursor_offset_2;
	/*
	 * If the user enable it, we also prepare for the compression
	 * framebuffer here.
	 */
	if (fPtr->compression != FALSE) {
		/*
		 * The compression buffer contains up to 256 bytes per line. If the
		 * real line can compress down to 256 bytes it is stored in this buffer
		 * and used for the next frame if the real line was not updated. The hardware
		 * detects a changed real line itself. To ensure this, line length must be
		 * 1024, 2048 or 4096 and the bits in GP_BLT_STATUS must be set accordingly.
		 * Every write into the framebuffer will be recalculated into a line with
		 * this setting. Every write sets in interal "dirty" bit for this line. The
		 * compressed buffer for this line will be updated the next time and (if
		 * successfull) this bit will be cleard again.
		 *
		 * How to place the compressed buffer:
		 * - behind the visual framebuffer (if space left)
		 * - between two visual lines if space left
		 */
		switch (fPtr->real_line_length) {
		case 1024:
		case 2048:
			fPtr->compression = 0;	/* TODO implement it, currently fall back only */
			break;
		case 4096:
			/* check free space between lines. Don't forget the hardware cursor image! */
			if ((fPtr->real_line_length - (bpp*width/8)) >= (0x100 + 0x100)) {
			/*
			 * we locate the compression buffer as follows:
			 * Note: v=visual line, c=compressed line, i=cursor image, .=free space
			 * [...]
			 * Line y-2: vvvvvvvvvvvvvvvvvvvvvvvvvccccccc..............
			 * Line y-1: vvvvvvvvvvvvvvvvvvvvvvvvvccccccc.........iiiii
			 * Line y  : vvvvvvvvvvvvvvvvvvvvvvvvvccccccc.........iiiii
			 *
			 * Note: Cursor image is only valid in the last two lines (see note above)
			 */
				fPtr->cbuffer_offset=bpp*width/8;
				fPtr->cbuffer_ldelta=4096;	/* same as visual line delta */
				fPtr->cbuffer_lsize=256;	/* could be smaller for smaller resolutions */
			}
			else
				fPtr->compression = 0;	/* not possible in this mode */
		}
	}
#if 0
	 /*
	  * TODO: Where is space for blit operations
	  * Problem is the SXGA mode with 16 colour depth
	  * It consumes the whole video memory due to 4096 bytes
	  * line offset. There is space in each line of about 1536 bytes.
	  * For the cursor its ok (256 Bytes) but for bitblit?
	  */
#endif
	return TRUE;
}
/**
 * @brief setup new video mode into the hardware
 * @param[in] pScrn screen
 * @param[in] mode new mode
 *
 * Set the new video mode. This is the job of the Geode GX1
 * framebuffer driver. We don't care here how to setup these
 * things.
 *
 * @return TRUE on success
 */
static Bool gx1SetNewVideoMode(ScrnInfoRec *pScrn, DisplayModeRec *mode)
{
	GeodeDevRec * fPtr = GEODEDEVPTR(pScrn);
	struct geodefb_ioctl geode_fb;
	int line_length;

	TRACE_ENTER(__FUNCTION__);
	/*
	 * framebuffer framework uses another type of
	 * mode definitions. Recalculate it
	 */
	xfree2fbdev_fblayout(pScrn, &geode_fb.si);
	xfree2fbdev_timing(mode,  &geode_fb.si);

#ifdef DEBUG
	print_xfree_mode("init",mode);
	print_fbdev_mode("init",&geode_fb.si);
#endif
	/*
	 * use acceleration as the user like
	 */
	geode_fb.enable_accel = fPtr->accel;
	if (geode_fb.enable_accel) {
		/*
		 * TODO: We are using the old real_line_length of the current mode here!
		 * So we may request a wrong scratch pad ram size!
		 * Example:
		 * Current mode 1280x1024@8 -> 2048
		 * New mode 1280x1024@16 -> 4096, but we request for 2048!!!
		 * FIXME: Move to the acceleration source part!
		 */
		line_length = (pScrn->bitsPerPixel>>3) * mode->HDisplay;
		/*
		 * it seems for the most common modes (XGA and SXGA) we only
		 * can use 2048 and 4096 scratch pad RAM size. See "curious"
		 * note in geode_accel.c
		 */
		if (line_length < 2049)
			geode_fb.sram_size=2048;
#if 0
		else if (line_length < 3073)
			geode_fb.sram_size=3072;	/* seems not usable */
#endif
		else
			geode_fb.sram_size=4096;

		xf86DrvMsg(pScrn->scrnIndex, X_INFO,"Calling for %d scratch size\n",geode_fb.sram_size);
	}
	else
		geode_fb.sram_size=0;

	/*
	 * set the mode
	 */
	if (0 != ioctl(fPtr->fd,FBIOSET_VMODE,(void*)(&geode_fb))) {
		xf86DrvMsg(pScrn->scrnIndex, X_ERROR,"FBIOSET_VMODE: %s\n", strerror(errno));
		return FALSE;
	}

	fPtr->real_line_length=geode_fb.line_length;
	xf86DrvMsg(pScrn->scrnIndex, X_INFO,"FBIOSET_VMODE: Line length now %d bytes\n", fPtr->real_line_length);
	fPtr->scratch_offset=geode_fb.sram_offset;
	/*
	 * if the scratch pad is enabled, we must set the
	 * BLT buffer bases also
	 */
	if (geode_fb.enable_accel) {
		geode_setup_blitbuffer_pointers(fPtr,geode_fb.sram_size,geode_fb.sram_offset);
	}
	xf86DrvMsg(pScrn->scrnIndex, X_INFO,"FBIOSET_VMODE: New scratch pad offset: %d\n", fPtr->scratch_offset);

	/* read back FIXME remove this */
	if (0 != ioctl(fPtr->fd,FBIOGET_FSCREENINFO,(void*)(&fPtr->fix))) {
		xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
			   "FBIOGET_FSCREENINFO: %s\n", strerror(errno));
		return FALSE;
	}
	if (0 != ioctl(fPtr->fd,FBIOGET_VSCREENINFO,(void*)(&fPtr->var))) {
		xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
			   "FBIOGET_VSCREENINFO: %s\n", strerror(errno));
		return FALSE;
	}
	/*
	 * calculate some offsets for acceleration
	 */
	geode_calc_offscreen_memories(pScrn->scrnIndex,fPtr,mode->HDisplay,mode->VDisplay,pScrn->bitsPerPixel);

	if (fPtr->compression != FALSE)
		geode_enable_compression(fPtr);

	return TRUE;
}

/**
 * @brief FIXME
 * @param[in] scrnIndex Specfies the screen index value.
 * @param[in] pMode pointer to the mode structure.
 * @param[in] flags may be used for status check?.
 *
 * This function will switches the screen mode
 *
 * What does it switch?
 * @return ?
 */
static Bool gx1SwitchMode(int scrnIndex, DisplayModeRec *mode, int flags)
{
	ScrnInfoRec *pScrn = xf86Screens[scrnIndex];
	GeodeDevRec * fPtr = GEODEDEVPTR(pScrn);

	TRACE_ENTER(__FUNCTION__);

	xfree2fbdev_timing(mode, &fPtr->var);
	if (0 != ioctl(fPtr->fd,FBIOPUT_VSCREENINFO,(void*)(&fPtr->var))) {
		xf86DrvMsg(scrnIndex, X_ERROR,
			   "FBIOPUT_VSCREENINFO: %s\n", strerror(errno));
		return FALSE;
	}
	return TRUE;
}

/**
 * @brief Switch back to previous mode
 * @param scrnIndex Specfies the screen index value.
 * @param flags Not used inside the function.
 *
 * This is called when VT switching back to console mode.
 *
 * @returns TRUE on success, FALSE anywhere else
 */
static void gx1LeaveVT(int scrnIndex, int flags)
{
	ScrnInfoPtr pScrn = xf86Screens[scrnIndex];

	TRACE_ENTER(__FUNCTION__);
/*
 * Seems video mode restore will be done by the
 * framebuffer console itself. We are in luck!
 */
	geode_restore_HW_state(pScrn);
}

/**
 * @brief Reenter the X screen
 * @param scrnIndex Specfies the screen index value.
 * @param flags Not used inside the function.
 *
 * This is called when VT switching back to the X server
 *
 * @returns TRUE on success, FALSE anywhere else
 */
static Bool gx1EnterVT(int scrnIndex, int flags)
{
	ScrnInfoPtr pScrn = xf86Screens[scrnIndex];

	TRACE_ENTER(__FUNCTION__);
	if (!gx1SetNewVideoMode(pScrn, pScrn->currentMode)) {
		xf86DrvMsg(scrnIndex, X_ERROR,"Setting video mode failes\n");
		return FALSE;
	}
/* We do not support panning yet */
/* fbdevHWAdjustFrame( scrnIndex, 0, 0, 0 ); */
/* fbdevHWAdjustFrame(scrnIndex, pScrn->frameX0, pScrn->frameY0, 0); */
	return TRUE;
}

/**
 * @brief Free all claimed resources of one screen
 * @param[in] scrnIndex FIXME
 * @param[in] pScreen FIXME
 *
 * Note: This function will be called whenever the X server ends
 * and also when only the current session ends only (restart of a screen).
 *
 * @return FIXME
 **/
static Bool FBDevCloseScreen( int scrnIndex, ScreenRec *pScreen )
{
	ScrnInfoRec *pScrn = xf86Screens[ scrnIndex ];
	GeodeDevRec *fPtr = GEODEDEVPTR( pScrn );

	TRACE_ENTER(__FUNCTION__);

	/*
	 * clean up various resources from the infrastructure
	 */
	geode_hw_cursor_exit(pScreen);
	if ( fPtr->accel )
		geode_exit_acceleration(pScreen);
#if 0
	if ( fPtr->pDGAMode ) {
		xfree( fPtr->pDGAMode );
		fPtr->pDGAMode = NULL;
		fPtr->nDGAMode = 0;
	}
#endif
	pScrn->vtSema = FALSE;

	/* restore hardware if we did some really ugly things */
	geode_restore_HW_state( pScrn );

	/* free all mapped resources */
	geode_unmap_IO_memory( pScrn );

	/* call the previous (saved) close function */
	pScreen->CloseScreen = fPtr->CloseScreen;
	return ( *pScreen->CloseScreen ) ( scrnIndex, pScreen );
}

/**
 * @brief seems to used when screen is rotated
 * @param[in] index
 * @param[in] x
 * @param[in] y
 **/
#if 0
static void FBDevPointerMoved( int index, int x, int y )
{
	ScrnInfoPtr pScrn = xf86Screens[ index ];
	GeodeDevRec * fPtr = GEODEDEVPTR( pScrn );
	int newX, newY;

	TRACE_ENTER(__FUNCTION__);
	switch ( fPtr->rotate ) {
	case FBDEV_ROTATE_CW:
		/* 90 degrees CW rotation. */
		newX = pScrn->pScreen->height - y - 1;
		newY = x;
		break;

	case FBDEV_ROTATE_CCW:
		/* 90 degrees CCW rotation. */
		newX = y;
		newY = pScrn->pScreen->width - x - 1;
		break;

	case FBDEV_ROTATE_UD:
		/* 180 degrees UD rotation. */
		newX = pScrn->pScreen->width - x - 1;
		newY = pScrn->pScreen->height - y - 1;
		break;

	default:
		/* No rotation. */
		newX = x;
		newY = y;
		break;
	}

	TRACE_EXIT(__FUNCTION__);
	/* Pass adjusted pointer coordinates to wrapped PointerMoved function. */
	( *fPtr->PointerMoved ) ( index, newX, newY );
}
#endif

/**
 * @brief FIXME
 * @param[in] pScrn
 * @param[in] op
 * @param[in] ptr
 **/
Bool FBDevDriverFunc( ScrnInfoPtr pScrn, xorgDriverFuncOp op, pointer ptr )
{
	xorgHWFlags * flag;

	TRACE_ENTER(__FUNCTION__);
	switch ( op ) {
	case GET_REQUIRED_HW_INTERFACES:
		flag = ( CARD32* ) ptr;
		( *flag ) = 0;
		return TRUE;
	}

	TRACE_EXIT(__FUNCTION__);
	return FALSE;
}

/**
 * @brief check if a given videomode is supported by the hardware
 * @param[in] scrnIndex Specfies the screen index value.
 * @param[in] pMode Pointer to the screen mode structure..
 * @param[in] verbose not used for implementation.
 * @param[in] flags not used for implementation
 *
 * This function checks if a mode is suitable for the selected chipset.
 * Used when VidMode extension is available and by xf86ValidateModes()
 *
 * @return MODE_OK if the specified mode is supported or MODE_BAD if not
 */
static ModeStatus gx1ValidMode(int scrnIndex,  DisplayModeRec *pMode, Bool Verbose, int flags)
{
#if 0
	typedef struct _DisplayModeRec {
		struct _DisplayModeRec *	prev;
		struct _DisplayModeRec *	next;
		char *			name;		/* identifier for the mode */
		ModeStatus			status;
		int				type;

		/* These are the values that the user sees/provides */
		int				Clock;		/* pixel clock freq (kHz) */
		int				HDisplay;	/* horizontal timing */
		int				HSyncStart;
		int				HSyncEnd;
		int				HTotal;
		int				HSkew;
		int				VDisplay;	/* vertical timing */
		int				VSyncStart;
		int				VSyncEnd;
		int				VTotal;
		int				VScan;
		int				Flags;

		/* These are the values the hardware uses */
		int				ClockIndex;
		int				SynthClock;	/* Actual clock freq to
								* be programmed  (kHz) */
		int				CrtcHDisplay;
		int				CrtcHBlankStart;
		int				CrtcHSyncStart;
		int				CrtcHSyncEnd;
		int				CrtcHBlankEnd;
		int				CrtcHTotal;
		int				CrtcHSkew;
		int				CrtcVDisplay;
		int				CrtcVBlankStart;
		int				CrtcVSyncStart;
		int				CrtcVSyncEnd;
		int				CrtcVBlankEnd;
		int				CrtcVTotal;
		Bool			CrtcHAdjusted;
		Bool			CrtcVAdjusted;
		int				PrivSize;
		INT32 *			Private;
		int				PrivFlags;

		float			HSync, VRefresh;
	} DisplayModeRec, *DisplayModePtr;
#endif
/* most of the modes we support */
#if 0
	ScrnInfoPtr pScrn = xf86Screens[scrnIndex];
	GeodeDevRec * fPtr = GEODEDEVPTR(pScrn);
	struct fb_var_screeninfo var;

	TRACE_ENTER(__FUNCTION__);

	xf86DrvMsg(scrnIndex, X_INFO,
		"Mode to check:\n"
		" users clock %dkHz\n"
		" hardware clock %dkHz\n"
		" users hdisplay %d pixels\n"
		" hardware hdisplay %d pixels\n"
		" users vdisplay %d lines\n"
		" hardware vdisplay %d lines\n",
		pMode->Clock,pMode->SynthClock,pMode->HDisplay,pMode->CrtcHDisplay,pMode->VDisplay,pMode->CrtcVDisplay);

	memcpy(&var,&fPtr->var,sizeof(var));
	xfree2fbdev_timing(pMode, &var);
	var.activate = FB_ACTIVATE_TEST;
	if (0 != ioctl(fPtr->fd,FBIOPUT_VSCREENINFO,(void*)(&fPtr->var))) {
		xf86DrvMsg(scrnIndex, X_ERROR,
			   "FBIOPUT_VSCREENINFO: %s\n", strerror(errno));
		TRACE("Mode is bad\n");
		return MODE_BAD;
	}
	TRACE("Mode is ok\n");
#endif
	return MODE_OK;
}

/**
 * @brief Invalidate video modes not suitable by this hardware
 * @param[inout] pScreenInfo blabla
 *
 * xf86ValidateModes will check that the mode HTotal and VTotal values
 * don't exceed the chipset's limit if pScreenInfo->maxHValue adn
 * pScreenInfo->maxVValue are set. Since our GX1ValidMode()
 * already takes care of this, we don't worry about setting them here.
 *
 */
static int gx1CheckVideoModes(ScrnInfoRec *pScreenInfo)
{
	/* graphics hardware features (not monitor!) */
	static ClockRange GeodeClockRange =
	{
		.minClock=25175,
		.maxClock=135000,
		.clockIndex=-1,
		.interlaceAllowed=FALSE,
		.doubleScanAllowed=TRUE,
		.ClockMulFactor=1,
		.ClockDivFactor=1,
		.PrivFlags=0
	};
	unsigned int PitchInc=32, minPitch = 320, maxPitch = 2048;
	unsigned int minHeight = 128, maxHeight = 1024;
	char **modes;
	int i;

/*	if (pScreenInfo->depth == 16)
		PitchInc = 2048;
	else
		PitchInc = 1024;

	PitchInc <<= 3; */	/* in bits */

	/*
	 * by default use what user sets in the XF86Config file
	 * Or the default modes if nothing was specified by the user
	 */
	modes = pScreenInfo->display->modes;

	/*
	 * this test will check all automatically registered
	 * video modes against the given monitor limits and
	 * rejects all modes that fails due to this. For modes
	 * possible for the monitor it calls our registed gx1ValidMode()
	 * to check if the graphics hardware is also able to handle
	 * this mode.
	 */
	i = xf86ValidateModes(pScreenInfo,
			      pScreenInfo->monitor->Modes,
			      modes,&GeodeClockRange,
			      NULL, minPitch, maxPitch,
			      PitchInc, minHeight, maxHeight,
			      pScreenInfo->display->virtualX,
			      pScreenInfo->display->virtualY,
			      4096*1024, /* FIXME */
			      LOOKUP_BEST_REFRESH);

	if (i == -1)
		return FALSE;
	/*
	* Prune the modes marked as invalid
	*/
	xf86PruneDriverModes(pScreenInfo);
	/*
	* at least one mode available?
	*/
	if (i == 0 || pScreenInfo->modes == NULL)
		return FALSE;

	xf86SetCrtcForModes(pScreenInfo, 0);	/* FIXME: ???? */
	/*
	* Set the current mode to the first in the list
	*/
	pScreenInfo->currentMode = pScreenInfo->modes;
	/*
	* Print the list of modes being used
	*/
	xf86PrintModes(pScreenInfo);

	return TRUE;
}

/**
 * @brief set the overscan colour
 * @param[in] pScrn FIXME
 * @param[in] overscan Overscan colour
 **/
static void geode_set_overscan(ScrnInfoRec *pScrn, int overscan)
{
	TRACE_ENTER(__FUNCTION__);
}


/**
 * @brief Init the screen and activate the requested video mode
 * @param[in] scrnIndex Specfies the screenindex value during generation.
 * @param[in] pScreen Pointer to screen info structure.
 * @param[in] argc parameters for command line arguments count
 * @param[in] argv command line arguments if any it is not used
 *
 * This function gets called at the start of each server generation
 *
 * Note: Per-generation data should be allocated with
 * AllocateScreenPrivateIndex() from the ScreenInit() function.
 *
 * @return ?
 */
static Bool FBDevScreenInit( int scrnIndex, ScreenPtr pScreen, int argc, char **argv )
{
	ScrnInfoPtr pScrn = xf86Screens[ pScreen->myNum ];
	GeodeDevRec * fPtr = GEODEDEVPTR( pScrn );
	VisualPtr visual;
	int width, height;

	TRACE_ENTER(__FUNCTION__);

#ifdef DEBUG
	ErrorF( "\tbitsPerPixel=%d, depth=%d, defaultVisual=%s\n"
	        "\tmask: %x,%x,%x, offset: %d,%d,%d\n",
	        pScrn->bitsPerPixel,
		pScrn->depth,
		xf86GetVisualName( pScrn->defaultVisual ),
		pScrn->mask.red, pScrn->mask.green, pScrn->mask.blue,
		pScrn->offset.red, pScrn->offset.green, pScrn->offset.blue );
#endif

	/*
	 * Map the videomemory itself
	 */
	xf86DrvMsg( scrnIndex, X_INFO, "Video Memory mapping\n" );
	if (geode_map_IO_memory(pScrn,fPtr) == FALSE ) {
		xf86DrvMsg( scrnIndex, X_ERROR, "mapping of video memory failed\n" );
		return FALSE;
	}

	/*
	 * this is the time to save the current hardware state. Maybe
	 * we will do strange things with the hardware the framebuffer driver
	 * can't handle
	 */
	xf86DrvMsg( scrnIndex, X_INFO, "Save current video mode\n" );
	geode_save_HW_state( pScrn );

	/*
	 * switch to the new video mode
	 */
	xf86DrvMsg( scrnIndex, X_INFO, "Set new video mode\n" );
	if ( !gx1SetNewVideoMode( pScrn, pScrn->currentMode ) ) {
		xf86DrvMsg( scrnIndex, X_ERROR, "mode initialization failed\n" );
		return FALSE;
	}
	/*
	 * we "own" the screen now
	 */
	pScrn->vtSema = TRUE;
	/*
	 * recalculate the line width in memory (in pixels)
	 * From now on "real_line_length" is valid!
	 */
	pScrn->displayWidth = fPtr->real_line_length / (pScrn->bitsPerPixel>>3);
	xf86DrvMsg( scrnIndex, X_INFO, "Line width is now %d pixels\n",pScrn->displayWidth);

	/* blank the screen (very late!) */
	xf86DrvMsg( scrnIndex, X_INFO, "Blank screen\n" );
	geode_save_screen( pScreen, SCREEN_SAVER_ON );

	/*
	 * Adjust the screen in case of panning
	 */
	xf86DrvMsg( scrnIndex, X_INFO, "Adjust frame\n" );
	fbdevHWAdjustFrame( scrnIndex, 0, 0, 0 );

	/*
	 * intitialise the mi layer
	 */
	xf86DrvMsg( scrnIndex, X_INFO, "MI (media independend) layer init\n" );
	miClearVisualTypes();	/* Re-initialise mi's visual list */
	if ( pScrn->bitsPerPixel > 8 ) {
		if ( !miSetVisualTypes( pScrn->depth, TrueColorMask, pScrn->rgbBits, TrueColor ) ) {
			xf86DrvMsg( scrnIndex, X_ERROR, "visual type setup failed for %d bits per pixel [1]\n", pScrn->bitsPerPixel );
			return FALSE;
		}
	} else {
		if ( !miSetVisualTypes( pScrn->depth, miGetDefaultVisualMask( pScrn->depth ), pScrn->rgbBits, pScrn->defaultVisual ) ) {
			xf86DrvMsg( scrnIndex, X_ERROR, "visual type setup failed for %d bits per pixel [2]\n", pScrn->bitsPerPixel );
			return FALSE;
		}
	}
	if ( !miSetPixmapDepths() ) {
		xf86DrvMsg( scrnIndex, X_ERROR, "pixmap depth setup failed\n" );
		return FALSE;
	}

	/* FIXME: Do we support rotation? */
	if ( fPtr->rotate == FBDEV_ROTATE_CW || fPtr->rotate == FBDEV_ROTATE_CCW ) {
		height = pScrn->virtualX;
		width = pScrn->displayWidth = pScrn->virtualY;
	} else {
		height = pScrn->virtualY;
		width = pScrn->virtualX;
	}
#if 0
	if ( fPtr->rotate && !fPtr->PointerMoved ) {
		fPtr->PointerMoved = pScrn->PointerMoved;
		pScrn->PointerMoved = FBDevPointerMoved;
	}
#endif
	fPtr->fbstart = fPtr->vmem_io + fPtr->fboff;
#ifdef DEBUG
	xf86DrvMsg( scrnIndex, X_INFO,
		    "Framebuffer Screen Init: \n"
				    " Videomemory start: %p\n"
				    " Width %d\n Heigth %d\n"
				    " DPI %d,%d\n Display width %d\n"
				    " BPP %d\n",
		    fPtr->fbstart, width, height,pScrn->xDpi,
		    pScrn->yDpi, pScrn->displayWidth, pScrn->bitsPerPixel );
#endif
	if (!fbScreenInit( pScreen, fPtr->fbstart, width, height, pScrn->xDpi, pScrn->yDpi, pScrn->displayWidth, pScrn->bitsPerPixel )) {
		xf86DrvMsg( scrnIndex, X_ERROR, "fbScreenInit failed!\n" );
		return FALSE;
	}

	if ( pScrn->bitsPerPixel > 8 ) {
		/* Fixup RGB ordering */
		visual = pScreen->visuals + pScreen->numVisuals;
		while ( --visual >= pScreen->visuals ) {
			if ( ( visual->class | DynamicClass ) == DirectColor ) {
				visual->offsetRed = pScrn->offset.red;
				visual->offsetGreen = pScrn->offset.green;
				visual->offsetBlue = pScrn->offset.blue;
				visual->redMask = pScrn->mask.red;
				visual->greenMask = pScrn->mask.green;
				visual->blueMask = pScrn->mask.blue;
			}
		}
	}

	/* must be after RGB ordering fixed */
	xf86DrvMsg( scrnIndex, X_INFO, "Picture init\n" );
	if ( !fbPictureInit( pScreen, NULL, 0 ) )
		xf86DrvMsg( pScrn->scrnIndex, X_WARNING,"Render extension initialisation failed\n" );

	/*
	 * Set initial black & white colourmap indices.
	 */
	xf86SetBlackWhitePixels( pScreen );

#if 0
	xf86DrvMsg( scrnIndex, X_INFO, "Init backing store\n" );
	/*
	 * If backing store is to be supported (as is usually the case),
	 * initialise it.
	 */
	miInitializeBackingStore( pScreen );	/* Whats it? */
	xf86SetBackingStore( pScreen );	/* Whats it? */
#endif

#if 0
	xf86DrvMsg( scrnIndex, X_INFO, "Shadow init\n" );
	if ( fPtr->shadowFB && !FBDevShadowInit( pScreen, fPtr ) ) {
		xf86DrvMsg( scrnIndex, X_ERROR,"shadow framebuffer initialization failed\n" );
		return FALSE;
	}
#endif
#if 0
	if ( !fPtr->rotate )  {
		xf86DrvMsg( scrnIndex, X_INFO, "DGA init\n" );
		FBDevDGAInit( pScrn, pScreen );
}
	else {
		xf86DrvMsg( scrnIndex, X_INFO, "display rotated; disabling DGA\n" );
		xf86DrvMsg( scrnIndex, X_INFO, "using driver rotation; disabling XRandR\n" );
		xf86DisableRandR();
}
#endif
	/*
	 * initialise the hardware acceleration before hardware cursor!
	 */
	if (fPtr->accel) {
		if (geode_init_acceleration(pScreen) != TRUE)
			xf86DrvMsg(scrnIndex, X_ERROR,"Hardware acceleration initialization failed\n");
		/* FIXME Should it fail here, if the initialisation fails? */
	}

	/*
	 * Initialise the cursor. First the software cursor,
	 * then the hardware cursor
	 */
	xf86DrvMsg( scrnIndex, X_INFO, "Init cursor\n" );
	miDCInitialize( pScreen, xf86GetPointerScreenFuncs() );
	if (geode_hw_cursor_init(pScreen) != TRUE)
		xf86DrvMsg(scrnIndex, X_ERROR,"Hardware cursor initialization failed\n");

	/* colormap */
	if ( !miCreateDefColormap( pScreen ) ) {
		xf86DrvMsg( scrnIndex, X_ERROR,"internal error: miCreateDefColormap failed in FBDevScreenInit()\n" );
		return FALSE;
	}

	/*
	 * register palette handling routines only on demand,
	 * otherwise they will be called whenever a programs starts
	 * but without any effect
	 */
	if ( pScrn->bitsPerPixel == 8 ) {
		if ( !xf86HandleColormaps(pScreen,
				256,	/* of colour is the CLUT */
				6,	/* size of each RGB entry in this CLUT */
				fbdevHWLoadPalette,
				geode_set_overscan,	/* manage overscan colour */
				0 ) )
			return FALSE;
	}

#if 0
	if (fPtr->ShadowFB)
		ShadowFBInit(pScreen, ApmRefreshArea /*FIXME*/);
#endif
	/*
	 * register a function for DMPS
	 * This feature must be enabled in the X server at buildtime!
	 * Also an <Option "dpms" "1"> must be given in the
	 * monitor-section in the xorg configuration file
	 */
	if (xf86DPMSInit( pScreen, geode_DPMS_feature, 0 ) != TRUE)
		xf86DrvMsg( scrnIndex, X_ERROR, "DPMS init failed!\n" );

	/*
	 * Register our own screen saver
	 */
	pScreen->SaveScreen = geode_save_screen;

	/* Wrap the current CloseScreen function */
	fPtr->CloseScreen = pScreen->CloseScreen;
	pScreen->CloseScreen = FBDevCloseScreen;

	/* Report any unused options (only for the first generation) */
	if (serverGeneration == 1)
		xf86ShowUnusedOptions(pScrn->scrnIndex, pScrn->options);

	TRACE_EXIT(__FUNCTION__);
	return TRUE;
}

/**
 * @brief Check if all required information is present and valid
 * @param[inout] pScrn Screen Handle (one per screen)
 * @param flags flags may be used to check the probed one with config.
 *
 * This function is only called once at server startup, and
 * not at the start of each server generation.  This means that
 * only things that are persistent across server generations can
 * be initialised here.  xf86Screens[] is (pScrn is a pointer to one
 * of these).  Privates allocated using xf86AllocateScrnInfoPrivateIndex()
 * are too, and should be used for data that must persist across
 * server generations.
 *
 * Note: No hardware should be mapped here, or must be unmapped again
 * when this function is left.
 *
 * @return TRUE on success and FALSE on failure.
 */
Bool FBDevPreInit( ScrnInfoRec *pScrn, int flags )
{
	GeodeDevRec *fPtr;
	const char *s;
	Gamma zeros = {0.0, 0.0, 0.0};

	TRACE_ENTER(__FUNCTION__);

	if ( flags & PROBE_DETECT )
		return FALSE;

	/*
	 * Check the number of entities, and fail if it isn't one.
	 * We could only handle one device here
	 */
	if ( pScrn->numEntities != 1 )
		return FALSE;
	/*
	 * allocate private data
	 */
	fPtr = xnfcalloc( sizeof( GeodeDevRec ), 1 );
	if (fPtr == NULL) {
		xf86DrvMsg( pScrn->scrnIndex, X_ERROR,"Cannot allocate private data. Giving up\n");
		return FALSE;
	}
	pScrn->driverPrivate = fPtr;	/* save for future work */

	/* FIXME ab jetzt mu bei einem Fehler diese driverPrivate wieder freigegeben werden! */
	/* FIXME ?????????????? */
	/* confScreen->monitor kommt aus zugehrigen
	*     Section "Monitor"
	* und enthlt von dort alle Informationen
	*/
	pScrn->monitor = pScrn->confScreen->monitor;
	xf86DrvMsg( pScrn->scrnIndex, X_INFO,
			"\n monitor.id=%s,\n"
			" monitor.vendor=%s,\n"
			" monitor.model=%s,\n"
			" monitor.nHsync=%d,\n"
			" monitor.nVrefresh=%d,\n"
			" monitor.widthmm=%d,\n"
			" monitor.heightmm=%d\n"
			" monitor.Modes=%p\n"
			" monitor.Last=%p\n",
			pScrn->monitor->id,pScrn->monitor->vendor,pScrn->monitor->model,
			pScrn->monitor->nHsync,pScrn->monitor->nVrefresh,pScrn->monitor->widthmm,
			pScrn->monitor->heightmm,pScrn->monitor->Modes,pScrn->monitor->Last);

	/*
	 * ????
	 */
	fPtr->pEnt = xf86GetEntityInfo( pScrn->entityList[0] );

	/* Was ist das? */
	pScrn->racMemFlags = RAC_FB | RAC_COLORMAP | RAC_CURSOR | RAC_VIEWPORT;
	/* XXX Is this right?  Can probably remove RAC_FB */
	pScrn->racIoFlags = RAC_FB | RAC_COLORMAP | RAC_CURSOR | RAC_VIEWPORT;

	/*
	 * xf86RegisterResources mit NULL-Liste ruft auf:
	*   xf86GetResourcesImplicitly()
	*     und bei PCI-Bus:  GetImplicitPciResources
	* Reserviert etwas das PCI-spezifisch ist
	*/
	if ( fPtr->pEnt->location.type == BUS_PCI && xf86RegisterResources( fPtr->pEnt->index, NULL, ResExclusive ) ) {
		xf86DrvMsg( pScrn->scrnIndex, X_ERROR,"xf86RegisterResources() found resource conflicts\n" );
		goto on_error_fail;
	}

	/*
	 * collect data about our hardware
	 */
	if ( !geode_hw_init( pScrn, NULL, xf86FindOptionValue( fPtr->pEnt->device->options, GEODEDEV_DRIVER_NAME ) ) )
		goto on_error_fail;
	/*
	 * Setup the depth and bits per pixel we prefer
	 * Note: This functions uses command line and/or config options to do so
	 */
	if ( !xf86SetDepthBpp( pScrn,
				16,	/* driver's preferred depth */
				0,	/* pixmap's preferred depth -> use default */
				16,	/* framebuffer's preferred depth */
				NoDepth24Support ) )
		goto on_error_fail;

	/* check depth of default visual */
	if ((pScrn->depth != 8) && (pScrn->depth != 16)) {
		xf86DrvMsg(pScrn->scrnIndex, X_ERROR,"Given depth (%d) is not supported by this driver\n",
				   pScrn->depth);
		goto on_error_fail;
	}
	xf86PrintDepthBpp( pScrn );

	/*
	 * This must happen after pScrn->display has been set because
	 * xf86SetWeight references it: color weight
	 */
	if ( pScrn->depth > 8 ) {
		rgb zeros = { 0, 0, 0 };	/* The defaults are OK for us */
		if ( !xf86SetWeight( pScrn, zeros, zeros ) )
			goto on_error_fail;
		else {
			; /* FIXME check that weight returned is supported */
		}
	}

	/* visual init, -1 mean use server's default */
	if ( !xf86SetDefaultVisual( pScrn, -1 ) )
		goto on_error_fail;

	/* We don't currently support DirectColor at > 8bpp */
	if ( pScrn->depth > 8 && pScrn->defaultVisual != TrueColor ) {
		xf86DrvMsg( pScrn->scrnIndex, X_ERROR, "requested default visual (%s) is not supported at depth %d\n", xf86GetVisualName( pScrn->defaultVisual ), pScrn->depth );
		goto on_error_fail;
	}

	/* The gamma fields must be initialised when using the new cmap code */
	if ( !xf86SetGamma( pScrn, zeros ) ) {
		xf86DrvMsg( pScrn->scrnIndex, X_ERROR, "Cannot set Gamma values\n");
		goto on_error_fail;
	}

	pScrn->progClock = TRUE;	/* We support a programmable clock */
	pScrn->chipset = GEODEDEV_DRIVER_NAME;
	pScrn->videoRam = fPtr->fix.smem_len;	/* size of video memory in bytes */

	xf86DrvMsg( pScrn->scrnIndex, X_INFO, "hardware: %s (video memory: %dkB)\n", fPtr->fix.id, pScrn->videoRam / 1024 );

	/*
	 * Start to handle given options
	 * 1st Collect them
	 */
	xf86CollectOptions( pScrn, NULL );
	/*
	 * 2nd process them
	 */
	if ( !( fPtr->Options = xalloc( sizeof( FBDevOptions ) ) ) )
		goto on_error_fail;
	memcpy( fPtr->Options, FBDevOptions, sizeof( FBDevOptions ) );
	xf86ProcessOptions( pScrn->scrnIndex, fPtr->pEnt->device->options, fPtr->Options );

	/* do not use shadow framebuffer by default */
	fPtr->shadowFB = xf86ReturnOptValBool( fPtr->Options, OPTION_SHADOW_FB, FALSE );
	xf86MarkOptionUsedByName(fPtr->Options, "ShadowFB");
	/* do not use acceleration by default */
	fPtr->accel = xf86ReturnOptValBool( fPtr->Options, OPTION_ACCEL, FALSE );
	xf86MarkOptionUsedByName(fPtr->Options, "Accel");
	/* do not use compression by default */
	fPtr->compression = xf86ReturnOptValBool( fPtr->Options, OPTION_COMPRESS, FALSE );
	xf86MarkOptionUsedByName(fPtr->Options, "Compression");
#if 0
	/* rotation */
	fPtr->rotate = FBDEV_ROTATE_NONE;
	if ( ( s = xf86GetOptValString( fPtr->Options, OPTION_ROTATE ) ) ) {
		xf86MarkOptionUsedByName(fPtr->Options, "Rotate");
		if ( !xf86NameCmp( s, "CW" ) ) {
			fPtr->shadowFB = TRUE;
			fPtr->rotate = FBDEV_ROTATE_CW;
			xf86DrvMsg( pScrn->scrnIndex, X_CONFIG, "rotating screen clockwise\n" );
		} else if ( !xf86NameCmp( s, "CCW" ) ) {
			fPtr->shadowFB = TRUE;
			fPtr->rotate = FBDEV_ROTATE_CCW;
			xf86DrvMsg( pScrn->scrnIndex, X_CONFIG, "rotating screen counter-clockwise\n" );
		} else if ( !xf86NameCmp( s, "UD" ) ) {
			fPtr->shadowFB = TRUE;
			fPtr->rotate = FBDEV_ROTATE_UD;
			xf86DrvMsg( pScrn->scrnIndex, X_CONFIG, "rotating screen upside-down\n" );
		} else {
			xf86DrvMsg( pScrn->scrnIndex, X_CONFIG, "\"%s\" is not a valid value for Option \"Rotate\"\n", s );
			xf86DrvMsg( pScrn->scrnIndex, X_INFO, "valid options are \"CW\", \"CCW\" and \"UD\"\n" );
		}
	}
#endif
	/* select video modes */
	gx1CheckVideoModes(pScrn);

	if ( fPtr->shadowFB )
		pScrn->displayWidth = pScrn->virtualX;	/* ShadowFB handles this correctly */
	else {
		/*
		 * what displayWidth we should use here? The framebuffer driver
		 * will return its current line length. But it can be the correct
		 * length, but also not. The only relevant value will be the length
		 * of the mode X will set later in ScreenInit.
		 */
		pScrn->displayWidth = fPtr->real_line_length / ( fPtr->var.bits_per_pixel >> 3 );
	}
		xf86DrvMsg( pScrn->scrnIndex, X_INFO, "displayWidth set to %d bytes per line\n",pScrn->displayWidth);

	/*
	 * Set display resolution in DPI
	 * This function should calculate it from
	 * the screen size so we give 0
	 */
	xf86SetDpi( pScrn, 0, 0 );

	/*
	 * load remaining modules on request
	 * FIXME: Maybe load cfb for 8 bit modes and cfb16 for 16 bit modes instead of fb?
	 */
	if ( xf86LoadSubModule( pScrn, "fb" ) == NULL ) {
		xf86DrvMsg( pScrn->scrnIndex, X_ERROR,"Cannot load framebuffer module\n");
		goto on_error_fail;
	}
	xf86LoaderReqSymLists( fbSymbols, NULL );

	/* Load XAA if needed */
	if ( fPtr->accel ) {
		if (!xf86LoadSubModule(pScrn, "xaa")) {
			xf86DrvMsg( pScrn->scrnIndex, X_ERROR,"Cannot load XAA module\n");
			goto on_error_fail;
		}
		xf86LoaderReqSymLists(xaaSymbols, NULL);
	}

	/* Load ramdac module for hardware cursor support */
	if (!xf86LoadSubModule(pScrn, "ramdac")) {
		xf86DrvMsg( pScrn->scrnIndex, X_ERROR,"Cannot load RAM DAC module\n");
		goto on_error_fail;
	}
	xf86LoaderReqSymLists(ramdacSymbols, NULL);

#if 0
	/* Load shadow if needed */
	if ( fPtr->shadowFB ) {
		xf86DrvMsg( pScrn->scrnIndex, X_CONFIG, "using shadow framebuffer\n" );
		if ( !xf86LoadSubModule( pScrn, "shadow" ) ) {
			goto on_error_fail;
		}
		xf86LoaderReqSymLists( shadowSymbols, NULL );
	}
#endif

	TRACE_EXIT(__FUNCTION__);
	return TRUE;

on_error_fail:

	if (pScrn->driverPrivate != NULL) {
		xfree( pScrn->driverPrivate );
		pScrn->driverPrivate = NULL;
	}
	return FALSE;
}

/**
 * @brief Probe for hardware we support in this driver
 * @param[in] drv Driver list struct to allocate ScrnInfoRec
 * @param[in] flags (0, PROBE_DEFAULT, PROBE_DETECT)
 *
 * Probe for the device we support, if it is present in this system
 *
 * If flags is PROBE_DETECT: Is used if "-configure" or "-probe" command
 * line arguments are given and indicates to the Probe() function that
 * it should not configure the bus entities and that no xorg.conf
 * information is available.
 *
 * NOTE: The xf86DrvMsg() functions cannot be used from
 *       within this probe function unless driverName or name is set.
 *
 * This probe must register all non-relocatable resources at this stage
 * and must not change anything in the hardware
 */
Bool FBDevProbe( DriverPtr drv, int flags )
{
	int i;
	ScrnInfoRec *pScrn=NULL;
	GDevPtr *devSections;
	int numDevSections;
	int bus, device, func;
	char *dev;
	Bool foundScreen = FALSE;
	int entity;

	TRACE_ENTER(__FUNCTION__);

	/* For now, just bail out for PROBE_DETECT. -> FIXME */
	if ( flags & PROBE_DETECT ) {
		TRACE("Detection only. Exit\n");
		return FALSE;
	}
	/*
	* try to find a matching section in
	* the xorg.conf for this driver
	*/
	if ( ( numDevSections = xf86MatchDevice( GEODEDEV_DRIVER_NAME, &devSections ) ) <= 0 ) {
		TRACE("No Section found. Exit\n");
		return FALSE;
	}

	/*
	* walk through the sections to find our hardware
	*/
	for ( i = 0; i < numDevSections; i++ ) {
		/*
		* query for the framebuffer device node name
		*/
		dev = xf86FindOptionValue( devSections[i]->options, GEODEDEV_DRIVER_NAME );

		/*
		* query for the PCI triple, where our hardware should be present
		*/
		if ( devSections[i]->busID ) {
			if ( xf86ParsePciBusString( devSections[i]->busID, &bus, &device,&func ) ) {
				if ( !xf86CheckPciSlot( bus, device, func ) )
					continue;
			} else {
				TRACE("No Geode-GX1 detected. Exit\n");
				return FALSE;	/* FIXME */
			}
		}

		/*
		* try to open the framebuffer device node
		* dev could be NULL, the environment variable FRAMEBUFFER will be used instead
		*/
		if ( geode_HW_probe(dev) ) {
			/* XXX what about when there's no busID set? */

			entity = xf86ClaimPciSlot( bus, device, func, drv,0, devSections[ i ],TRUE );
			pScrn = xf86ConfigPciEntity( pScrn, 0, entity, NULL, RES_SHARED_VGA, NULL, NULL, NULL, NULL );
		}
		/*
		* fill the screen with all relevant information
		*/
		if ( pScrn ) {
			foundScreen = TRUE;

			pScrn->driverVersion = GEODEDEV_VERSION;
			/* xf86DrvMsg() can't be called until setting these two entries */
			pScrn->driverName = GEODEDEV_DRIVER_NAME;
			pScrn->name = GEODEDEV_NAME;
			pScrn->Probe = FBDevProbe;
			pScrn->PreInit = FBDevPreInit;
			pScrn->ScreenInit = FBDevScreenInit;
			pScrn->SwitchMode = gx1SwitchMode;
			pScrn->AdjustFrame = fbdevHWAdjustFrame;
			pScrn->EnterVT = gx1EnterVT;
			pScrn->LeaveVT = gx1LeaveVT;
			pScrn->ValidMode = gx1ValidMode;
			pScrn->FreeScreen = geode_free_screen;
			pScrn->EnableDisableFBAccess = NULL;
			pScrn->SetDGAMode = NULL;	/* wozu dient DGA eigentlich? */
			/* will not be used???? see xfree86/common/xf86cfmap.c. Seems to overwrite it! */
			pScrn->ChangeGamma = geode_change_gamma;
#		if 0
			pScrn->PointerMoved;
			pScrn->PMEvent;
			pScrn->HandleMessage;
#		endif
			/*
			 * is this correct? Once used in the xf86DPMSInit() call,
			 * second called here?? When?
			 */
			pScrn->DPMSSet = geode_DPMS_feature;	/* gets called when the DPMS level is changed */
#		if 1
			pScrn->LoadPalette = fbdevHWLoadPalette;
			pScrn->SetOverscan = geode_set_overscan;
#		endif
#		if 0
			pScrn->RRFunc = NULL;
#		endif

			xf86DrvMsg( pScrn->scrnIndex, X_CONFIG, "claimed PCI slot %d:%d:%d\n", bus, device, func );
			xf86DrvMsg( pScrn->scrnIndex, X_INFO,"using %s\n", dev ? dev : "default device" );
		}
	}
	xfree( devSections );
	TRACE_EXIT(__FUNCTION__);
	return foundScreen;
}

/* end of file geode-fb.c */
